"""
AlgorithmicBiasAnalyzer Module
==============================

This module provides the `AlgorithmicBiasAnalyzer` class, which is responsible for detecting
algorithmic bias and explainability issues in machine learning models. It includes functionalities
for:
- Categorical and continuous feature classification
- Predictive parity evaluation
- Equalized odds computation
- Explainability and bias evaluation
- Train-test bias evaluation
- Confusion matrix analysis
"""

import os
import seaborn as sns  
import configparser
import math
import matplotlib
matplotlib.use('Agg')  
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from sklearn.metrics import confusion_matrix, precision_score, recall_score, roc_auc_score, f1_score, accuracy_score

# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')

class AlgorithmicBiasAnalyzer:
    """
    AlgorithmicBiasAnalyzer Class
    =============================

    Detects algorithmic bias and explainability issues in machine learning models by analyzing
    various bias metrics and generating corresponding visualizations.

    Attributes
    ----------
    train : pd.DataFrame
        The training dataset.
    test : pd.DataFrame
        The testing dataset.
    __train_df : pd.DataFrame
        Cleaned training dataset with missing values dropped.
    __test_df : pd.DataFrame
        Cleaned testing dataset with missing values dropped.
    __model : sklearn-like model
        The machine learning model to analyze.
    columns : list
        List of column names from the training dataset.
    __algorithmic_biases : dict
        Dictionary to store detected algorithmic biases.
    __categorical_columns : list
        List of categorical columns to analyze for bias.
    upload_folder : str
        Path to the upload folder for saving generated images.

    Methods
    -------
    get_algorithmic_biases()
        Retrieves the detected algorithmic biases.
    __predict(data)
        Generates predictions using the provided model.
    describe_algo_bias_tests(results, test_name, algorithmic_biases=None)
        Generates descriptive information based on bias test results.
    compute_equalized_odds()
        Computes Equalized Odds metrics for specified categorical columns.
    evaluate_predictive_parity()
        Evaluates Predictive Parity for specified categorical columns.
    evaluate_feature_importances()
        Evaluates model explainability and bias based on feature importances.
    evaluate_confusion_matrix()
        Evaluates bias between train and test datasets using confusion matrices.
    """
    def __init__(self, train_file, test_file, model, categorical_columns, upload_folder):
        """
        Initialize the AlgorithmicBiasAnalyzer.

        Parameters
        ----------
        train_file : str
            Path to the training dataset CSV file.
        test_file : str
            Path to the testing dataset CSV file.
        model : sklearn-like model
            The machine learning model to analyze.
        categorical_columns : list
            List of categorical column names to analyze for bias.
        upload_folder : str
            Path to the upload folder for saving generated images.

        Raises
        ------
        ValueError
            If there is an error initializing the analyzer with the provided files or model.
        """
        try:
            self.train = pd.read_csv(train_file)
            self.test = pd.read_csv(test_file)
            self.__train_df = self.train.dropna()
            self.__test_df = self.test.dropna()
            self.__model = model
            self.columns = self.__train_df.columns.tolist()
            self.__algorithmic_biases = {"Equalized Odds": [], 
                                         "Predictive Parity": [], 
                                         "Explainability": [], 
                                         "Confusion Matrix": []}
            
            self.__categorical_columns = categorical_columns

            self.upload_folder = upload_folder

            if isinstance(self.__train_df, pd.DataFrame):
                self.train_columns = self.__train_df.columns.tolist()
            elif isinstance(self.columns, list):
                self.train_columns = self.columns
            else:
                self.train_columns = None

        except Exception as e:
            raise ValueError(f"Error initializing agent: {e}")

    def get_algorithmic_biases(self):
        """
        Retrieve the detected algorithmic biases.

        Returns
        -------
        dict
            A dictionary containing the detected algorithmic biases categorized by test type.
        """
        return self.__algorithmic_biases

    def __predict(self, data):
        """
        Generate predictions using the provided model.

        Parameters
        ----------
        data : pd.DataFrame or np.ndarray
            Dataset to generate predictions for.

        Returns
        -------
        np.ndarray
            Predicted labels.

        Raises
        ------
        RuntimeError
            If there is an error during prediction.
        """
        try:
            X = data[:, :-1] if isinstance(data, np.ndarray) else data.iloc[:, :-1]
            return self.__model.predict(X)
        except Exception as e:
            raise RuntimeError(f"Error during prediction: {e}")

    @staticmethod
    def describe_algo_bias_tests(results, test_name, algorithmic_biases=None):
        """
        Generate descriptive information based on bias test results.

        Parameters
        ----------
        results : dict
            The results from the bias tests.
        test_name : str
            The name of the bias test.
        algorithmic_biases : dict, optional
            Dictionary to store detailed bias descriptions.

        Returns
        -------
        dict
            The updated results dictionary with descriptive information added.
        """
        if test_name == 'equalized_odds':
            EQUALIZED_ODDS_THRESHOLD = float(config['thresholds']['EQUALIZED_ODDS_THRESHOLD'])
            DELTA_TPR_THRESHOLD = float(config['thresholds']['DELTA_TPR_THRESHOLD'])
            DELTA_FPR_THRESHOLD = float(config['thresholds']['DELTA_FPR_THRESHOLD'])

            for col, metrics in results.items():
                if "error" in metrics:
                    metrics["description"] = "Error encountered during calculation."
                    continue

                equalized_odds = metrics.get("Equalized Odds", None)
                delta_tpr = metrics.get("Delta TPR", None)
                delta_fpr = metrics.get("Delta FPR", None)
                tpr_fpr = metrics.get("TPR_FPR", None)

                description = []
                is_biased = False  

                # Equalized Odds
                if equalized_odds > EQUALIZED_ODDS_THRESHOLD:
                    description.append(f"Equalized odds value is {equalized_odds:.2f}, indicating significant disparity.")
                else:
                    description.append(f"Equalized odds value is {equalized_odds:.2f}, within acceptable limits.")

                # Delta TPR
                if delta_tpr > DELTA_TPR_THRESHOLD:
                    description.append(f"Delta TPR is {delta_tpr:.2f}, exceeding the threshold.")
                else:
                    description.append(f"Delta TPR is {delta_tpr:.2f}, indicating minimal variation.")

                # Delta FPR
                if delta_fpr > DELTA_FPR_THRESHOLD:
                    description.append(f"Delta FPR is {delta_fpr:.2f}, exceeding the threshold.")
                else:
                    description.append(f"Delta FPR is {delta_fpr:.2f}, indicating minimal variation.")

                # Conclusion
                if equalized_odds > EQUALIZED_ODDS_THRESHOLD:
                    description.append(
                        "Conclusion: Bias detected based on the evaluated metrics.")
                    is_biased = True
                else:
                    description.append(
                        "Conclusion: No significant bias detected based on the evaluated metrics.")


                metrics["description"] = description
                metrics["isBiased"] = is_biased

                # Append ethical question if biased and equalized odds > threshold
                if is_biased:
                    max_tpr_group = max(tpr_fpr, key=lambda g: tpr_fpr[g]["TPR"])
                    min_tpr_group = min(tpr_fpr, key=lambda g: tpr_fpr[g]["TPR"])
                    max_fpr_group = max(tpr_fpr, key=lambda g: tpr_fpr[g]["FPR"])
                    min_fpr_group = min(tpr_fpr, key=lambda g: tpr_fpr[g]["FPR"])


                    algorithmic_biases["Equalized Odds"].append(
                        f"The {col} column has an equalized odds value of {equalized_odds:.2f}, "
                        f"because the TPR difference between {max_tpr_group} ({tpr_fpr[max_tpr_group]['TPR']:.2f}) "
                        f"and {min_tpr_group} ({tpr_fpr[min_tpr_group]['TPR']:.2f}) is high, "
                        f"and the FPR difference between {max_fpr_group} ({tpr_fpr[max_fpr_group]['FPR']:.2f}) "
                        f"and {min_fpr_group} ({tpr_fpr[min_fpr_group]['FPR']:.2f}) is high."
                    )

            return results
        elif test_name == 'predictive_parity':
            for col, col_data in results.items():
                if "error" in col_data:
                    # Skip columns with errors
                    col_data["description"] = "Error encountered during calculation."
                    continue

                precision_diff = col_data.get("Precision Diff", 0)
                recall_diff = col_data.get("Recall Diff", 0)
                roc_diff = col_data.get("ROC Diff", 0)

                description = []
                is_biased = False  # Initialize the bias flag

                precision_threshold = float(config['thresholds']['precision_threshold_predictive_parity'])
                recall_threshold = float(config['thresholds']['recall_threshold_predictive_parity'])
                roc_threshold = float(config['thresholds']['roc_threshold_predictive_parity'])

                # Analyze Precision Diff
                if precision_diff > precision_threshold:
                    description.append(f"Precision Diff ({precision_diff:.2f}) exceeds the threshold ({precision_threshold:.2f}), indicating potential bias.")
                else:
                    description.append(f"Precision Diff ({precision_diff:.2f}) is within acceptable limits.")

                # Analyze Recall Diff
                if recall_diff > recall_threshold:
                    description.append(f"Recall Diff ({recall_diff:.2f}) exceeds the threshold ({recall_threshold:.2f}), indicating potential bias.")
                else:
                    description.append(f"Recall Diff ({recall_diff:.2f}) is within acceptable limits.")

                # Analyze ROC Diff
                if roc_diff > roc_threshold:
                    description.append(f"ROC Diff ({roc_diff:.2f}) exceeds the threshold ({roc_threshold:.2f}), indicating potential bias.")
                else:
                    description.append(f"ROC Diff ({roc_diff:.2f}) is within acceptable limits.")

                # Add a conclusion
                if precision_diff > precision_threshold or recall_diff > recall_threshold or roc_diff > roc_threshold:
                    description.append(
                        "Conclusion: Bias detected based on the evaluated metrics.")
                    is_biased = True
                else:
                    description.append(
                        "Conclusion: No significant bias detected based on the evaluated metrics.")

                # Update the results with the description
                col_data["description"] = description
                col_data["isBiased"] = is_biased

                if is_biased:
                    max_precision_group = max(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["Precision"])
                    min_precision_group = min(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["Precision"])
                    max_recall_group = max(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["Recall"])
                    min_recall_group = min(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["Recall"])
                    max_roc_group = max(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["ROC AUC"])
                    min_roc_group = min(col_data["Metrics"], key=lambda g: col_data["Metrics"][g]["ROC AUC"])

                    reasons = []

                    if precision_diff > precision_threshold:
                        reasons.append(f"precision of {max_precision_group} ({col_data['Metrics'][max_precision_group]['Precision']:.2f})and {min_precision_group} ({col_data['Metrics'][min_precision_group]['Precision']:.2f})")

                    if recall_diff > recall_threshold:
                        reasons.append(f"recall of {max_recall_group} ({col_data['Metrics'][max_recall_group]['Recall']:.2f}) and {min_recall_group} ({col_data['Metrics'][min_recall_group]['Recall']:.2f})")

                    if roc_diff > roc_threshold:
                        reasons.append(f"ROC AUC of {max_roc_group} ({col_data['Metrics'][max_roc_group]['ROC AUC']:.2f}) and {min_roc_group} ({col_data['Metrics'][min_roc_group]['ROC AUC']:.2f})")

                    bias_reason = (
                        f"The {col} column has a bias because the {', '.join(reasons)}."
                    )

                    algorithmic_biases["Predictive Parity"].append(bias_reason)


            return results
        elif test_name == 'explainability':
            try:
                description = []

                # 1. Explainability
                if results["Is Explainable"]:
                    description.append("The __model is explainable as it provides feature importance or coefficients.")
                else:
                    description.append("The __model is not explainable as it lacks feature importance or interpretability methods.")

                # 2. Transparency
                if results["Is Transparent"]:
                    description.append("The __model is transparent and shows how it makes decisions.")
                else:
                    description.append("The __model is not transparent and does not provide insight into its decision-making process.")

                # 3. Top 3 Features
                if results["Top 3 Features"]:
                    top_features_text = "Top 3 important features according to the __model are:\n"
                    for feature in results["Top 3 Features"]:
                        top_features_text += f"  - {feature['Feature Name']}: {feature['Importance Ratio']:.2f}\n"
                    description.append(top_features_text.strip())
                else:
                    description.append("No important features could be identified from the __model.")

                # 4. Reason (Bias Detection)
                if results["Bias Detected"]:
                    description.append("Bias detected: One of the top features dominates with ≥ 50% importance.")
                else:
                    description.append("No significant bias detected: Feature importance is more evenly distributed.")

                # 5. Conclusion
                if results["Is Explainable"] and results["Is Transparent"] and not results["Bias Detected"]:
                    description.append("Conclusion: The __model is explainable, transparent, and does not exhibit significant bias.")
                elif results["Bias Detected"]:
                    description.append("Conclusion: The __model exhibits significant bias in feature importance.")
                else:
                    description.append("Conclusion: The __model lacks either explainability or transparency, potentially raising concerns.")

                return description

            except Exception as e:
                raise RuntimeError(f"Error generating explainability description: {e}")
        elif test_name == 'confusion_matrix':
            try:
                train_metrics = results.get("Train", {})
                test_metrics = results.get("Test", {})
                performance_disparities = results.get("Performance Disparities", {})

                train_descriptions = []
                test_descriptions = []

                # Train Descriptions
                train_descriptions.append(
                    f"Train set accuracy is {train_metrics['Accuracy']:.2f}, precision is {train_metrics['Precision']:.2f}, recall is {train_metrics['Recall']:.2f}, and F1-score is {train_metrics['F1-Score']:.2f}."
                )
                
                accuracy_threshold = float(config['thresholds']['accuracy_confusion_matrix'])
                precision_threshold = float(config['thresholds']['precision_confusion_matrix'])
                recall_threshold = float(config['thresholds']['recall_confusion_matrix'])
                f1_threshold = float(config['thresholds']['f1_confusion_matrix'])

                # Check where the __model might lack in Train
                train_lack = []
                if train_metrics['Accuracy'] < accuracy_threshold:
                    train_lack.append("accuracy")
                if train_metrics['Precision'] < precision_threshold:
                    train_lack.append("precision")
                if train_metrics['Recall'] < recall_threshold:
                    train_lack.append("recall")
                if train_metrics['F1-Score'] < f1_threshold:
                    train_lack.append("F1-score")

                if train_lack:
                    train_descriptions.append(f"The __model lacks in train {' and '.join(train_lack)}.")
                else:
                    train_descriptions.append("The __model performs well in all evaluated metrics for the train set.")

                # Check underfitting condition
                underfitting = any(
                    train_metrics[metric] < 0.7 and test_metrics[metric] < 0.7
                    for metric in ["Accuracy", "Precision", "Recall", "F1-Score"]
                )

                if underfitting:
                    train_descriptions.append("The __model might be underfitting as both train and test metrics are below 0.7 for one or more metrics.")
                else:
                    train_descriptions.append("The __model is not underfitting as train and test metrics are sufficiently high.")


                # Train Conclusion
                if any(value > 0.2 for value in performance_disparities.values()) or any(
                    train_metrics[metric] < 0.7 for metric in ["Accuracy", "Precision", "Recall", "F1-Score"]
                ):
                    train_descriptions.append("Conclusion: The __model exhibits bias in the train set due to performance disparities or underperforming metrics.")
                else:
                    train_descriptions.append("Conclusion: The __model does not exhibit significant bias in the train set.")


                # Test Descriptions
                test_descriptions.append(
                    f"Test set accuracy is {test_metrics['Accuracy']:.2f}, precision is {test_metrics['Precision']:.2f}, recall is {test_metrics['Recall']:.2f}, and F1-score is {test_metrics['F1-Score']:.2f}."
                )

                # Check where the __model might lack in Test
                test_lack = []
                if test_metrics['Accuracy'] < accuracy_threshold:
                    test_lack.append("accuracy")
                if test_metrics['Precision'] < precision_threshold:
                    test_lack.append("precision")
                if test_metrics['Recall'] < recall_threshold:
                    test_lack.append("recall")
                if test_metrics['F1-Score'] < f1_threshold:
                    test_lack.append("F1-score")

                if test_lack:
                    test_descriptions.append(f"The __model lacks in test {' and '.join(test_lack)}.")
                else:
                    test_descriptions.append("The __model performs well in all evaluated metrics for the test set.")

                overfitting = any(value > 0.2 for value in performance_disparities.values())

                if overfitting:
                    test_descriptions.append("The __model is overfitting as one or more performance disparities exceed 0.2.")
                else:
                    test_descriptions.append("The __model is not overfitting as performance disparities are within acceptable limits.")

                # Test Conclusion
                if overfitting or any(
                    test_metrics[metric] < 0.7 for metric in ["Accuracy", "Precision", "Recall", "F1-Score"]):
                    test_descriptions.append("Conclusion: The __model might exhibit some bias in the test set based on performance disparities.")
                else:
                    test_descriptions.append("Conclusion: The __model does not exhibit significant bias in the test set.")

                # Check for performance disparities > 0.1 and append ethical question
                for metric, disparity in performance_disparities.items():
                    if disparity > 0.2:
                        algorithmic_biases["Confusion Matrix"].append(
                            f"The performance disparity for {metric} is {disparity:.2f}, "
                            f"with Train {metric}: {train_metrics[metric]:.2f} and Test {metric}: {test_metrics[metric]:.2f}."
                        )

                # Append descriptions to results
                if "description" not in train_metrics:
                    train_metrics["description"] = []
                if "description" not in test_metrics:
                    test_metrics["description"] = []

                train_metrics["description"].extend(train_descriptions)
                test_metrics["description"].extend(test_descriptions)

                return results
            except Exception as e:
                raise RuntimeError(f"Error generating descriptive information: {e}")

    def compute_equalized_odds(self):
        """
        Compute Equalized Odds metrics for specified categorical columns.

        Parameters
        ----------
        None

        Returns
        -------
        dict
            A dictionary containing Equalized Odds metrics, descriptive information, and paths to generated images.

        Raises
        ------
        RuntimeError
            If there is an error during the computation of Equalized Odds.
        """

        try:
            categorical_columns = self.__categorical_columns
            if not categorical_columns and self.train_columns is None:
                raise RuntimeError("No column names provided, unable to compute Equalized Odds.")
            
            if not self.__model:
                raise RuntimeError("Sorry, No information regarding the Equalized Odds as no __model exists.")

            # If user didn't provide anything, we attempt default columns [age, gender, race]
            if not categorical_columns:
                default_cols = ["age", "gender", "race"]
                if self.train_columns is None:
                    analyze_columns = default_cols
                else:
                    analyze_columns = [c for c in self.train_columns if any(dc in c.lower() for dc in default_cols)]
                    if not analyze_columns:
                        raise RuntimeError("No relevant default columns found for subgroup analysis.")
            else:
                provided_cols = categorical_columns
                if not provided_cols:
                    raise RuntimeError("No valid categorical columns provided.")
                if self.train_columns is None:
                    if not isinstance(self.__test_df, pd.DataFrame):
                        raise RuntimeError("No column names provided and data is not a DataFrame, cannot proceed.")
                    for col in provided_cols:
                        if col not in self.__test_df.columns:
                            raise RuntimeError(f"Column '{col}' not found in test data.")
                    analyze_columns = provided_cols
                else:
                    if not isinstance(self.__test_df, pd.DataFrame):
                        raise RuntimeError("Test data is not a DataFrame, cannot find columns by name.")
                    for col in provided_cols:
                        if col not in self.__test_df.columns:
                            raise RuntimeError(f"Column '{col}' not found in test data.")
                    analyze_columns = provided_cols

            # Check if any column has too many unique values
            for col in analyze_columns:
                unique_vals = self.__test_df[col].unique()
                if len(unique_vals) > 10:
                    raise RuntimeError(
                        f"Column '{col}' has more than 10 unique values, likely continuous. Cannot compute Equalized Odds."
                    )

            # Get predictions and ground truth
            predictions = self.__predict(self.__test_df)
            true_labels = self.__test_df.iloc[:, -1] if isinstance(self.__test_df, pd.DataFrame) else None
            if true_labels is None:
                raise RuntimeError("Test data does not have labels in the last column.")

            results = {}

            for col in analyze_columns:
                subgroups = self.__test_df[col].unique()
                tpr_fpr = {}

                # Calculate TPR & FPR for each subgroup
                for group in subgroups:
                    idx = self.__test_df[col] == group
                    y_true_group = true_labels[idx]
                    y_pred_group = predictions[idx]

                    if len(y_true_group) == 0:
                        continue

                    try:
                        tn, fp, fn, tp = confusion_matrix(y_true_group, y_pred_group, labels=[0,1]).ravel()
                    except ValueError:
                        continue

                    tpr = tp / (tp + fn) if (tp + fn) > 0 else 0.0
                    fpr = fp / (fp + tn) if (fp + tn) > 0 else 0.0
                    tpr_fpr[group] = {"TPR": tpr, "FPR": fpr}

                if len(tpr_fpr) < 2:
                    results[col] = {
                        "error": f"The column '{col}' has only one subgroup which is insufficient for comparison."
                    }
                    continue

                # Calculate Delta TPR, Delta FPR, Equalized Odds
                subg = list(tpr_fpr.keys())
                delta_tpr = max(
                    abs(tpr_fpr[a]["TPR"] - tpr_fpr[b]["TPR"])
                    for a in subg for b in subg if a != b
                )
                delta_fpr = max(
                    abs(tpr_fpr[a]["FPR"] - tpr_fpr[b]["FPR"])
                    for a in subg for b in subg if a != b
                )
                equalized_odds = delta_tpr + delta_fpr

               
                sns.set_style("whitegrid")
                groups = list(tpr_fpr.keys())
                tpr_values = [tpr_fpr[group]["TPR"] for group in groups]
                fpr_values = [tpr_fpr[group]["FPR"] for group in groups]

                # Prepare data for Seaborn
                plot_data = []
                for i, grp in enumerate(groups):
                    plot_data.append({"Group": grp, "RateType": "TPR", "Value": tpr_values[i]})
                    plot_data.append({"Group": grp, "RateType": "FPR", "Value": fpr_values[i]})

                df_plot = pd.DataFrame(plot_data)

                plt.figure(figsize=(8, 6))
                barplot = sns.barplot(
                    x="Group", 
                    y="Value", 
                    hue="RateType", 
                    data=df_plot, 
                    palette="Set2"
                )
                barplot.set_title(f"TPR and FPR by Subgroup ({col})", fontsize=14, pad=15)
                barplot.set_xlabel("Subgroups", fontsize=12)
                barplot.set_ylabel("Rates", fontsize=12)
                barplot.tick_params(axis='x', rotation=45)
                barplot.legend(title="Rate Type")
                plt.tight_layout()

                image_filename = f"{col}_equalized_odds.png"
                image_path = os.path.join(self.upload_folder, image_filename)
                plt.savefig(image_path)
                plt.close()

                results[col] = {
                    "TPR_FPR": tpr_fpr,
                    "Delta TPR": delta_tpr,
                    "Delta FPR": delta_fpr,
                    "Equalized Odds": equalized_odds,
                    "image": image_filename 
                }

            # Call the static method to generate descriptions
            results_desc = self.describe_algo_bias_tests(results=results, test_name='equalized_odds', algorithmic_biases=self.__algorithmic_biases)
            def convert_keys_to_str(obj):
                if isinstance(obj, dict):
                    return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
                elif isinstance(obj, list):
                    return [convert_keys_to_str(i) for i in obj]
                else:
                    return obj
            return convert_keys_to_str(results_desc)

        except Exception as e:
            raise RuntimeError(f"Error computing Equalized Odds: {e}")

    def evaluate_predictive_parity(self):
        """
            Evaluate Predictive Parity for specified categorical columns.

            Parameters
            ----------
            None

            Returns
            -------
            dict
                A dictionary containing Predictive Parity metrics, descriptive information, and paths to generated images.

            Raises
            ------
            RuntimeError
                If there is an error during the evaluation of Predictive Parity.
            """
        try:
            categorical_columns = self.__categorical_columns
            if not categorical_columns and self.train_columns is None:
                raise RuntimeError("No column names provided, unable to evaluate Predictive Parity.")

            if not self.__model:
                raise RuntimeError("Sorry, No information regarding the Predictive Parity as no __model exists.")

            if not categorical_columns:
                default_cols = ["age", "gender", "race"]
                if self.train_columns is None:
                    analyze_columns = default_cols
                else:
                    analyze_columns = [c for c in self.train_columns if any(dc in c.lower() for dc in default_cols)]
                    if not analyze_columns:
                        raise RuntimeError("No relevant default columns found for subgroup analysis.")
            else:
                provided_cols = categorical_columns
                if not provided_cols:
                    raise RuntimeError("No valid categorical columns provided.")
                if self.train_columns is None:
                    if not isinstance(self.__test_df, pd.DataFrame):
                        raise RuntimeError("No column names provided and data is not a DataFrame, cannot proceed.")
                    for col in provided_cols:
                        if col not in self.__test_df.columns:
                            raise RuntimeError(f"Column '{col}' not found in test data.")
                    analyze_columns = provided_cols
                else:
                    if not isinstance(self.__test_df, pd.DataFrame):
                        raise RuntimeError("Test data is not a DataFrame, cannot find columns by name.")
                    for col in provided_cols:
                        if col not in self.__test_df.columns:
                            raise RuntimeError(f"Column '{col}' not found in test data.")
                    analyze_columns = provided_cols

            for col in analyze_columns:
                unique_vals = self.__test_df[col].unique()
                if len(unique_vals) > 10:
                    raise RuntimeError(f"Column '{col}' has more than 10 unique values, likely continuous. Cannot compute Predictive Parity.")

            predictions = self.__predict(self.__test_df)
            if isinstance(self.__test_df, pd.DataFrame):
                true_labels = self.__test_df.iloc[:, -1]
            else:
                raise RuntimeError("Test data does not have labels in the last column.")

            if hasattr(self.__model, "predict_proba"):
                probabilities = self.__model.predict_proba(self.__test_df.iloc[:, :-1])[:, 1]
            else:
                raise RuntimeError("__model does not support probability predictions for ROC AUC computation.")

            results = {}

            categories = ["Precision", "Recall", "ROC AUC"]
            N = len(categories)
            angles = [n / float(N) * 2 * math.pi for n in range(N)]
            angles += angles[:1] 

            for col in analyze_columns:
                subgroups = self.__test_df[col].unique()
                if len(subgroups) == 0:
                    results[col] = {"error": f"The column '{col}' has only one subgroup which is insufficient for comparison."}
                    continue

                metrics = {}
                for group in subgroups:
                    idx = self.__test_df[col] == group
                    y_true_group = true_labels[idx]
                    y_pred_group = predictions[idx]
                    y_prob_group = probabilities[idx]
                    if len(y_true_group) == 0:
                        continue

                    precision = precision_score(y_true_group, y_pred_group, zero_division=1)
                    recall = recall_score(y_true_group, y_pred_group, zero_division=1)
                    roc = roc_auc_score(y_true_group, y_prob_group) if len(np.unique(y_true_group)) > 1 else None

                    metrics[group] = {
                        "Precision": precision,
                        "Recall": recall,
                        "ROC AUC": (roc if roc is not None else 0.0)
                    }

                if len(metrics) < 2:
                    results[col] = {"error": f"The column '{col}' has only one subgroup which is insufficient for comparison."}
                    continue

                # compute diffs
                precision_values = [m["Precision"] for m in metrics.values()]
                recall_values = [m["Recall"] for m in metrics.values()]
                roc_values = [m["ROC AUC"] for m in metrics.values()]

                precision_diff = max(precision_values) - min(precision_values) if precision_values else 0
                recall_diff = max(recall_values) - min(recall_values) if recall_values else 0
                roc_diff = max(roc_values) - min(roc_values) if roc_values else 0

                # Plot Radar for each column
                plt.figure(figsize=(7, 7))
                plt.title(f"Predictive Parity Radar ({col})", size=14, y=1.08)

                # For each subgroup, we create a line on the spider
                subgroup_names = list(metrics.keys())
                for group in subgroup_names:
                    # get values in the order of [Precision, Recall, ROC AUC], then close the circle
                    val_list = [
                        metrics[group]["Precision"],
                        metrics[group]["Recall"],
                        metrics[group]["ROC AUC"]
                    ]
                    val_list.append(val_list[0])

                    plt.polar(angles, val_list, label=group)
                    plt.fill(angles, val_list, alpha=0.1)  

  
                plt.xticks(angles[:-1], categories)
                plt.yticks([0.0, 0.25, 0.5, 0.75, 1.0], ["0.0","0.25","0.5","0.75","1.0"], color="grey", size=8)
                plt.ylim(0,1)

                plt.legend(loc='upper right', bbox_to_anchor=(1.3, 1.0))

                image_filename = f"{col}_predictive_parity.png"
                image_path = os.path.join(self.upload_folder, image_filename)
                plt.savefig(image_path, bbox_inches='tight')
                plt.close()

                results[col] = {
                    "Metrics": metrics,
                    "Precision Diff": precision_diff,
                    "Recall Diff": recall_diff,
                    "ROC Diff": roc_diff,
                    "image": image_filename
                }

            results_desc = self.describe_algo_bias_tests(results=results, test_name='predictive_parity', algorithmic_biases=self.__algorithmic_biases)
            
            def convert_keys_to_str(obj):
                if isinstance(obj, dict):
                    return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
                elif isinstance(obj, list):
                    return [convert_keys_to_str(i) for i in obj]
                else:
                    return obj
            return convert_keys_to_str(results_desc)

        except Exception as e:
            raise RuntimeError(f"Error evaluating Predictive Parity: {e}")

    def evaluate_feature_importances(self):
        """
        Evaluate model explainability and bias based on feature importances.

        Parameters
        ----------
        None

        Returns
        -------
        dict
            A dictionary containing explainability metrics, bias detection results, and paths to generated images.

        Raises
        ------
        RuntimeError
            If there is an error during the evaluation of feature importances.
        """
        try:
            if not self.__model:
                raise RuntimeError("Sorry, No information regarding the Explainability as no __model exists.")

            if not self.__model or not (hasattr(self.__model, "feature_importances_") or hasattr(self.__model, "coef_")):
                return {
                    "Is Explainable": False,
                    "Is Transparent": False,
                    "Bias Detected": True,
                    "Reason": "__model lacks feature importance or interpretability methods.",
                    "Top 3 Features": None,
                    "images": []
                }

            # Get feature importances or coefficients
            if hasattr(self.__model, "feature_importances_"):
                feature_importances = self.__model.feature_importances_
            elif hasattr(self.__model, "coef_"):
                feature_importances = np.abs(self.__model.coef_).flatten()
            else:
                raise ValueError("__model does not provide feature importance or coefficients.")

            # Normalize them to sum to 1
            feature_importances = feature_importances / np.sum(feature_importances)

            # Get feature names
            if hasattr(self.__model, "feature_names_in_"):
                feature_names = self.__model.feature_names_in_
            elif isinstance(self.__train_df, pd.DataFrame):
                feature_names = self.__train_df.columns[:-1]  # Exclude the last column (target)
            else:
                feature_names = [f"Feature_{i}" for i in range(len(feature_importances))]

            # Get top 3 features
            top_features_idx = np.argsort(feature_importances)[-3:][::-1]
            top_features = [
                {"Feature Name": feature_names[idx], "Importance Ratio": float(feature_importances[idx])}
                for idx in top_features_idx
            ]

            # Compute "Others" percentage
            top_features_importance = sum(feature_importances[idx] for idx in top_features_idx)
            others_importance = 1 - top_features_importance


            labels = [f["Feature Name"] for f in top_features] + ["Others"]
            values = [f["Importance Ratio"] for f in top_features] + [others_importance]

            values = [v * 100 for v in values]

            cmap = LinearSegmentedColormap.from_list("importance_cmap", ["lightblue", "lightgreen", "lightcoral"])
            colors = [cmap(i / len(labels)) for i in range(len(labels))]

            plt.figure(figsize=(7, 7))
            plt.pie(
                values,
                labels=labels,
                colors=colors,
                autopct='%1.1f%%',
                startangle=140,
                pctdistance=0.85  
            )
            plt.title("Top 3 Feature Importances with Others", fontsize=14, pad=20)
            plt.tight_layout()
            image_filename = "explainability_bias_pie_updated.png"
            plt.savefig(os.path.join(self.upload_folder, image_filename))
            plt.close()

            feature_importance_threshold = float(config['thresholds']['feature_importance_threshold'])
            bias_detected = any(f["Importance Ratio"] >= feature_importance_threshold for f in top_features)

            results = {
                "Is Explainable": True,
                "Is Transparent": True,
                "Bias Detected": bias_detected,
                "Reason": ("One of the top 3 features dominates (≥ 95%)" if bias_detected
                        else "Feature importance is more evenly distributed."),
                "Top 3 Features": top_features,
                "images": [image_filename]
            }

            desc = self.describe_algo_bias_tests(results=results, test_name='explainability')
            results["description"] = desc

            if results["Bias Detected"]:
                explainable_status = "explainable" if results["Is Explainable"] else "not explainable"
                transparency_status = "transparent" if results["Is Transparent"] else "not transparent"
                
                # Format the top 3 features with their names and percentages
                top_features = results["Top 3 Features"]
                features_text = ", ".join(
                    f"{feature['Feature Name']} ({feature['Importance Ratio'] * 100:.2f}%)"
                    for feature in top_features
                )
                
                # Construct the sentence
                bias_sentence = (
                    f"The __model is {explainable_status}, transparency is {transparency_status}, "
                    f"and the __model's top 3 features are: {features_text}."
                )
                
                # Append this to self.algo_ethical_questions['Explainability']
                self.__algorithmic_biases["Explainability"].append(bias_sentence)
            def convert_keys_to_str(obj):
                if isinstance(obj, dict):
                    return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
                elif isinstance(obj, list):
                    return [convert_keys_to_str(i) for i in obj]
                else:
                    return obj
            return convert_keys_to_str(results)

        except Exception as e:
            raise RuntimeError(f"Error evaluating explainability and bias: {e}")

    def evaluate_confusion_matrix(self):
        """
            Evaluate bias between train and test datasets using confusion matrices.

            Parameters
            ----------
            None

            Returns
            -------
            dict
                A dictionary containing train and test performance metrics, performance disparities,
                descriptive information, and paths to generated confusion matrix images.

            Raises
            ------
            RuntimeError
                If there is an error during the evaluation of train-test bias.
            """
        try:
            if self.__train_df is None or self.__test_df is None:
                raise RuntimeError("Train or test data is not provided.")

            if not self.__model:
                raise RuntimeError("Sorry, No information regarding the Confusion Matrix as no __model exists.")

            datasets = {"Train": self.__train_df, "Test": self.__test_df}
            performance_metrics = {}
            confusion_matrix_images = {}

            # Evaluate performance for Train & Test
            for dataset_name, dataset in datasets.items():
                X = dataset.iloc[:, :-1]
                y_true = dataset.iloc[:, -1]
                y_pred = self.__model.predict(X)

                accuracy = accuracy_score(y_true, y_pred)
                precision = precision_score(y_true, y_pred, zero_division=1)
                recall = recall_score(y_true, y_pred, zero_division=1)
                f1 = f1_score(y_true, y_pred, zero_division=1)

                performance_metrics[dataset_name] = {
                    "Accuracy": accuracy,
                    "Precision": precision,
                    "Recall": recall,
                    "F1-Score": f1
                }

                # Generate confusion matrix
                cm = confusion_matrix(y_true, y_pred)
                plt.figure(figsize=(6, 4))
                sns.heatmap(
                    cm,
                    annot=True,
                    fmt="d",
                    cmap="Blues",
                    xticklabels=["Negative", "Positive"],
                    yticklabels=["Negative", "Positive"]
                )
                plt.title(f"{dataset_name} Confusion Matrix")
                plt.xlabel("Predicted")
                plt.ylabel("Actual")
                plt.tight_layout()

                image_filename = f"{dataset_name.lower()}_confusion_matrix.png"
                image_path = os.path.join(self.upload_folder, image_filename)
                plt.savefig(image_path, dpi=300)
                plt.close()

                confusion_matrix_images[dataset_name] = image_filename

            # Calculate performance disparities
            performance_disparities = {
                metric: abs(performance_metrics["Train"][metric] - performance_metrics["Test"][metric])
                for metric in performance_metrics["Train"]
            }

            # Prepare final output
            final_output = {
                "Train": {
                    **performance_metrics["Train"],
                    "Confusion Matrix Image": confusion_matrix_images["Train"]
                },
                "Test": {
                    **performance_metrics["Test"],
                    "Confusion Matrix Image": confusion_matrix_images["Test"]
                },
                "Performance Disparities": performance_disparities
            }

            results_desc = self.describe_algo_bias_tests(results=final_output, 
                                                         test_name='confusion_matrix', 
                                                         algorithmic_biases=self.__algorithmic_biases)
            def convert_keys_to_str(obj):
                if isinstance(obj, dict):
                    return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
                elif isinstance(obj, list):
                    return [convert_keys_to_str(i) for i in obj]
                else:
                    return obj
            return convert_keys_to_str(results_desc)

        except Exception as e:
            raise RuntimeError(f"Error evaluating train-test bias: {e}")
