"""
DataBiasAnalyzer Module
=======================

This module provides the `DataBiasAnalyzer` class, which is responsible for analyzing bias
in datasets by comparing training and testing datasets. It includes functionalities for:
- Categorical skewness detection
- Continuous skewness detection
- Outlier detection
- Disparate impact analysis
- Comparison of train and test datasets using statistical tests
"""

import os
import configparser
import numpy as np
import matplotlib
matplotlib.use('Agg')  
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.stats import skew
import matplotlib.pyplot as plt
from scipy.stats import chi2_contingency



config = configparser.RawConfigParser()
config.read('config.ini')

class DataBiasAnalyzer:
    """
    DataBiasAnalyzer Class
    ======================

    Analyzes bias in datasets by comparing training and testing datasets across various
    dimensions, including categorical skewness, continuous skewness, outlier detection,
    disparate impact, and comparison of train and test datasets using statistical tests.

    Attributes
    ----------
    train : pd.DataFrame
        The training dataset.
    test : pd.DataFrame
        The testing dataset.
    __train_df : pd.DataFrame
        Cleaned training dataset with missing values dropped.
    __test_df : pd.DataFrame
        Cleaned testing dataset with missing values dropped.
    __data_biases : dict
        Dictionary to store detected data biases across different tests.
    upload_folder : str
        Path to the upload folder for saving generated images.
    train_columns : list or None
        List of column names from the training dataset, if available.
    __categorical_columns : list
        List of categorical columns to analyze for bias.
    __continuous_columns : list
        List of continuous columns to analyze for bias.

    Methods
    -------
    compute_categorical_columns()
        Identifies and returns categorical columns based on unique value thresholds.
    compute_continuous_columns()
        Identifies and returns continuous columns based on unique value thresholds.
    get_data_biases()
        Retrieves the detected data biases.
    describe_data_bias_tests(results, test_name, dataset=None)
        Generates descriptive information based on data bias test results.
    calculate_categorical_skewness()
        Calculates categorical skewness for specified columns in training and testing datasets.
    calculate_continuous_skewness()
        Calculates continuous skewness for specified columns in training and testing datasets.
    detect_plot_outliers()
        Detects and visualizes outliers for specified columns in training and testing datasets.
    calculate_disparate_impact()
        Performs disparate impact analysis for specified categorical columns.
    compare_train_test()
        Compares train and test datasets for categorical bias using Chi-Square tests.
    """
    def __init__(self, train_file, test_file, upload_folder):
        """Initialize the DataBiasAnalyzer.

        Parameters
        ----------
        train_file : str
            Path to the CSV file containing the training dataset.
        test_file : str
            Path to the CSV file containing the testing dataset.
        upload_folder : str
            Path to the upload folder for saving generated images.

        Raises
        ------
        ValueError
            If there is an error initializing the analyzer with the provided files.
        """
        self.train = pd.read_csv(train_file)
        self.test = pd.read_csv(test_file)
        self.__train_df = self.train.dropna()
        self.__test_df = self.test.dropna()
        self.__data_biases = {"Categorical Skewness": [], "Continous Skewness": [], "Outlier Detection": [], "Disparate Impact": [], "Compare Train Test": []}

        self.upload_folder = upload_folder

        if isinstance(self.__train_df, pd.DataFrame):
            self.train_columns = self.__train_df.columns.tolist()
        else:
            self.train_columns = None

        self.__categorical_columns = self.compute_categorical_columns()
        self.__continuous_columns=self.compute_continuous_columns()

    def compute_categorical_columns(self):
        """Identify and return categorical columns based on unique value thresholds.

        Returns
        -------
        list
            List of categorical columns.

        Raises
        ------
        ValueError
            If the train or test data is not a valid pandas DataFrame.
        """
        if not isinstance(self.__train_df, pd.DataFrame):
            raise ValueError("Train data is not a valid pandas DataFrame.")

        if not isinstance(self.__test_df, pd.DataFrame):
            raise ValueError("Test data is not a valid pandas DataFrame.")

        # Extract feature names, excluding the last column as ground truth
        feature_names = self.__train_df.columns[:-1]

        # Ensure all train features exist in test dataset
        missing_features = [
            feature for feature in feature_names if feature not in self.__test_df.columns
        ]
        if missing_features:
            raise ValueError(f"The following features from train data are missing in test data: {missing_features}")

        categorical_cols = []
        for feature in feature_names:
            unique_values = self.__train_df[feature].nunique()
            if unique_values <= 10:
                categorical_cols.append(feature)

        return categorical_cols

    def compute_continuous_columns(self):
        """Identify and return continuous columns based on unique value thresholds.

        Returns
        -------
        list
            List of continuous columns.

        Raises
        ------
        ValueError
            If the train or test data is not a valid pandas DataFrame.
        """
        if not isinstance(self.__train_df, pd.DataFrame):
            raise ValueError("Train data is not a valid pandas DataFrame.")

        if not isinstance(self.__test_df, pd.DataFrame):
            raise ValueError("Test data is not a valid pandas DataFrame.")

        # Extract feature names, excluding the last column as ground truth
        feature_names = self.__train_df.columns[:-1]

        # Ensure all train features exist in test dataset
        missing_features = [
            feature for feature in feature_names if feature not in self.__test_df.columns
        ]
        if missing_features:
            raise ValueError(f"The following features from train data are missing in test data: {missing_features}")

        continuous_cols = []
        for feature in feature_names:
            unique_values = self.__train_df[feature].nunique()
            if unique_values > 10:
                continuous_cols.append(feature)

        return continuous_cols

    def get_data_biases(self):
        """Retrieve the detected data biases.

        Returns
        -------
        dict
            A dictionary containing detected data biases categorized by test type.
        """
        return self.__data_biases
    
    @staticmethod
    def describe_data_bias_tests(results, test_name, dataset=None):
        """Generate descriptive information based on data bias test results.

        Parameters
        ----------
        results : dict
            The results from the data bias tests.
        test_name : str
            The name of the data bias test.
        dataset : pd.DataFrame, optional
            The dataset being analyzed (e.g., 'Train' or 'Test').

        Returns
        -------
        dict
            The updated results dictionary with descriptive information added.

        Raises
        ------
        RuntimeError
            If there is an error generating the descriptive information.
        """
        def generate_categorical_description(column_name, dataset_name, column_data):
                """
            Generate descriptions for a single column in a given dataset (Train/Test).

            Parameters
            ----------
            column_name : str
                The name of the column being described.
            dataset_name : str
                Either 'Train' or 'Test'.
            column_data : dict
                The data for the column, including Bias and Subgroups.

            Returns
            -------
            list
                Two strings - the detailed description and the concise conclusion.
            """
                if column_data["isBiased"]:
                    majority_subgroup = max(column_data["Subgroups"], key=column_data["Subgroups"].get)
                    minority_subgroup = min(column_data["Subgroups"], key=column_data["Subgroups"].get)
                    majority_value = column_data["Subgroups"][majority_subgroup]
                    minority_value = column_data["Subgroups"][minority_subgroup]

                    detailed_description = (
                        f"The column '{column_name}' in the {dataset_name} dataset shows bias because its majority proportion "
                        f"subgroup '{majority_subgroup}' (proportion: {majority_value:.2f}) is more than 1.5 times larger than the "
                        f"minority subgroup '{minority_subgroup}' (proportion: {minority_value:.2f})."
                    )
                    concise_conclusion = (
                        f"Conclusion: '{column_name}' in {dataset_name} is biased: majority subgroup '{str(int(majority_subgroup))}' is dominant."
                    )
                else:
                    detailed_description = (
                        f"The column '{column_name}' in the {dataset_name} dataset does not show bias. All subgroups have "
                        f"proportions within acceptable limits."
                    )
                    concise_conclusion = (
                        f"Conclusion: '{column_name}' in {dataset_name} is not biased: subgroup proportions are balanced."
                    )

                return [detailed_description, concise_conclusion]
        def generate_disparate_description(ratios, dataset_name, column_name):
            """
            Generates descriptions for a specific column in a dataset based on DI ratios.
            """
            descriptions = []
            bias_exists = False

            disparate_impact_threshold = float(config['thresholds']['disparate_impact_threshold'])
            for outcome, subgroups in ratios.items():
                biased_subgroups = [
                    f"'{subgroup}' (DI ratio: {di_ratio:.2f})"
                    for subgroup, di_ratio in subgroups.items() if di_ratio < disparate_impact_threshold
                ]
                str_subgroups = ", ".join(biased_subgroups)

                if biased_subgroups:
                    bias_exists = True
                    descriptions.append(
                        f"In the {dataset_name} dataset, for the column '{column_name}', outcome '{outcome}', "
                        f"the data shows bias as the following subgroups have DI ratios less than 0.2: {', '.join(biased_subgroups)}."
                    )
                else:
                    descriptions.append(
                        f"In the {dataset_name} dataset, for the column '{column_name}', outcome '{outcome}', "
                        f"all subgroups have DI ratios greater than or equal to 0.2, indicating no bias."
                    )

            # Add a conclusion based on whether bias exists
            if bias_exists:
                descriptions.append(
                    f"Conclusion: The column '{column_name}' in the {dataset_name} dataset exhibits bias as at {str_subgroups} subgroup(s) hasvea DI ratio less than 0.2."
                )
            else:
                descriptions.append(
                    f"Conclusion: The column '{column_name}' in the {dataset_name} dataset does not exhibit bias, as all subgroups have DI ratios greater than or equal to 0.2."
                )

            return descriptions, bias_exists
        
        
        if test_name == 'categorical_skewness':
            for dataset_name, dataset_key in [("Train", "Train_bias_columns"), ("Test", "Test_bias_columns")]:
                for column_name, column_data in results[dataset_key].items():
                    descriptions = generate_categorical_description(column_name, dataset_name, column_data)
                    column_data["description"] = descriptions

            return results

        elif test_name == 'disparate_impact':
            for column, ratios in results.get("train_disparate_impact", {}).items():
                descriptions, is_biased = generate_disparate_description(ratios, "Train", column)
                results["train_disparate_impact"][column]["description"] = descriptions
                results["train_disparate_impact"][column]["isBiased"] = is_biased

            for column, ratios in results.get("test_disparate_impact", {}).items():
                descriptions, is_biased = generate_disparate_description(ratios, "Test", column)
                results["test_disparate_impact"][column]["description"] = descriptions
                results["test_disparate_impact"][column]["isBiased"] = is_biased

            return results
    
        elif test_name == 'continuous_skewness':
            for dataset in ['train', 'test']:
                if dataset in results:
                    for column, column_data in results[dataset].items():
                        skew_value = column_data.get('skewness', 0)

                        if skew_value > 0:
                            skew_description = (
                                f"The column '{column}' in the {dataset} dataset is left skewed with a value of {skew_value:.2f}. "
                                "High skewness values (>2 or <-2) indicate significant skewness."
                            )
                        else:
                            skew_description = (
                                f"The column '{column}' in the {dataset} dataset is right skewed with a value of {skew_value:.2f}. "
                                "High skewness values (>2 or <-2) indicate significant skewness."
                            )

                        is_biased = True if abs(skew_value) > 2 else False

                        if abs(skew_value) > 2:
                            conclusion = (
                                f"The column '{column}' exhibits a significant bias in the {dataset} dataset due to its high skewness value. "
                                "Consider normalizing or transforming this column for better model performance."
                            )
                        else:
                            conclusion = (
                                f"The column '{column}' does not exhibit significant bias in the {dataset} dataset based on skewness."
                            )

                        column_data['description'] = [skew_description, conclusion]
                        column_data['isBiased'] = is_biased
            return results
        
        elif test_name == 'outliers':
            for column, data in results.items():
                total_values_count = dataset[column].shape[0]
                outliers_count = data["count"]


                # Generate description
                description = (
                    f"The column '{column}' has {round((outliers_count/total_values_count)*100,2)}% outliers "
                )
                is_biased = False

                # Generate conclusion
                outlier_ratio = outliers_count / total_values_count
                if outlier_ratio > 0.1:  
                    conclusion = (
                        f"Conclusion: The column '{column}'is biased due to "
                        f"a high proportion of outliers ({outliers_count} outliers, {outlier_ratio:.2%} of total values)."
                    )
                    is_biased = True
                else:
                    conclusion = (
                        f"Conclusion: The column '{column}' has a low proportion of outliers. "
                        f"The dataset does not exhibit bias in this column."
                    )

                # Add descriptions to the results
                data["description"] = [description, conclusion]
                data["isBiased"] = is_biased

        elif test_name == 'compare_train_test':
            for _, stats in results.items():
                description = []
                is_biased = False

                # First bullet: Proportions similarity based on p-value
                if stats['p_value'] < 0.05:
                    description.append("The proportions of train and test are significantly different (p < 0.05).")
                else:
                    description.append("The proportions of train and test are not significantly different (p >= 0.05).")


                # Second bullet: Overall conclusion
                if stats['p_value'] < 0.05:
                    description.append("There may be a potential bias in this feature's distribution between train and test datasets.")
                    is_biased = True
                else:
                    description.append("The feature's distribution appears consistent between train and test datasets.")

                # Add the description to the results
                stats['description'] = description
                stats["isBiased"] = is_biased

    def calculate_categorical_skewness(self):
        """Calculate categorical skewness for specified columns in training and testing datasets.

        Returns
        -------
        dict
            Results dictionary with bias detection and associated visualizations.

        Raises
        ------
        RuntimeError
            If there is an error during the calculation of categorical skewness.
        """

        categorical_columns = self.__categorical_columns
        def convert_numpy_types(obj):
            """
            Recursively convert numpy types to native Python types.
            """
            if isinstance(obj, dict):
                return {k: convert_numpy_types(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_numpy_types(i) for i in obj]
            elif isinstance(obj, (np.bool_, bool)):
                return bool(obj)
            elif isinstance(obj, (np.integer, int)):
                return int(obj)
            elif isinstance(obj, (np.floating, float)):
                return float(obj)
            else:
                return obj

        if self.train_columns is None:
            raise RuntimeError("No column names provided, unable to evaluate Categorical Skewness.")

        # Figure out which columns to analyze
        if categorical_columns is None:
            default_cols = ["age", "gender", "race"]
            analyze_columns = [c for c in self.train_columns if any(dc in c.lower() for dc in default_cols)]
            if not analyze_columns:
                raise RuntimeError("No default sensitive columns (age, gender, race) found for categorical skewness.")
        else:
            provided_cols = categorical_columns
            if not provided_cols:
                raise RuntimeError("No valid categorical columns provided for categorical skewness.")
            for col in provided_cols:
                if col not in self.train.columns or col not in self.__test_df.columns:
                    raise RuntimeError(f"Column '{col}' not found in train/test data.")
            analyze_columns = provided_cols

        # Ensure columns appear truly categorical (≤10 unique subgroups)
        for col in analyze_columns:
            if self.__train_df[col].nunique() > 10:
                raise RuntimeError(
                    f"Column '{col}' has more than 10 unique values, likely continuous. "
                    "Cannot evaluate categorical skewness."
                )

        def detect_bias(proportions):
            """
            Detect biased subgroups based on proportions.
            """
            if proportions.empty or len(proportions) < 2:
                return False, {}

            sorted_vals = proportions.sort_values(ascending=True)
            minority_val = sorted_vals.iloc[0]
            majority_val = sorted_vals.iloc[-1]

            # Check if bias exists
            bias_exists = (majority_val - minority_val) > 1.5 * minority_val
            biased_subgroups = {
                subgroup: value for subgroup, value in proportions.items()
            }
            return bias_exists, biased_subgroups

        train_bias_columns = {}
        test_bias_columns = {}
        images = []

        sns.set_style("whitegrid")

        for column in analyze_columns:
            train_proportions = self.__train_df[column].value_counts(normalize=True).sort_index()
            bias_train, biased_subgroups_train = detect_bias(train_proportions)

            train_bias_columns[column] = {
                "isBiased": bias_train,
                "Subgroups": biased_subgroups_train,
                "Image": None
            }


            bar_colors = []
            for subgroup in train_proportions.index:
                if not bias_train:
                    bar_colors.append("lightgray")
                elif subgroup == train_proportions.idxmax():  
                    bar_colors.append("#FF6F61")  
                elif subgroup == train_proportions.idxmin():  
                    bar_colors.append("#58D68D")  
                else:
                    bar_colors.append("#AED6F1")

            plt.figure(figsize=(8, 5))
            ax = sns.barplot(
                x=train_proportions.index,
                y=train_proportions.values,
                palette=bar_colors
            )

            x_positions = range(len(train_proportions.index))
            ax.plot(
                x_positions,
                train_proportions.values,
                marker="o",
                color="black",
                label="Proportion"
            )
            ax.set_xticks(x_positions) 
            ax.set_xticklabels(train_proportions.index)  
            ax.set_title(f"Train Proportions for '{column}'", fontsize=12)
            ax.set_xlabel("Subgroups", fontsize=10)
            ax.set_ylabel("Proportion", fontsize=10)
            ax.legend()
            plt.xticks(rotation=45)
            plt.tight_layout()

            train_image_filename = f"{column}_train_categorical_skewness.png"
            train_image_path = os.path.join(self.upload_folder, train_image_filename)
            plt.savefig(train_image_path)
            plt.close()

            train_bias_columns[column]["Image"] = train_image_filename
            images.append(train_image_filename)

            test_proportions = self.__test_df[column].value_counts(normalize=True).sort_index()
            bias_test, biased_subgroups_test = detect_bias(test_proportions)

            test_bias_columns[column] = {
                "isBiased": bias_test,
                "Subgroups": biased_subgroups_test,
                "Image": None
            }

            bar_colors_test = []
            for subgroup in test_proportions.index:
                if not bias_test:
                    bar_colors_test.append("lightgray")
                elif subgroup == test_proportions.idxmax(): 
                    bar_colors_test.append("#FF6F61")  # Red for majority
                elif subgroup == test_proportions.idxmin():  
                    bar_colors_test.append("#58D68D")  # Green for minority
                else:
                    bar_colors_test.append("#AED6F1") 

            plt.figure(figsize=(8, 5))
            ax2 = sns.barplot(
                x=test_proportions.index,
                y=test_proportions.values,
                palette=bar_colors_test
            )

            x_positions_test = range(len(test_proportions.index))
            ax2.plot(
                x_positions_test,
                test_proportions.values,
                marker="o",
                color="black",
                label="Proportion"
            )
            ax2.set_xticks(x_positions_test)  
            ax2.set_xticklabels(test_proportions.index)  
            ax2.set_title(f"Test Proportions for '{column}'", fontsize=12)
            ax2.set_xlabel("Subgroups", fontsize=10)
            ax2.set_ylabel("Proportion", fontsize=10)
            ax2.legend()
            plt.xticks(rotation=45)
            plt.tight_layout()

            test_image_filename = f"{column}_test_categorical_skewness.png"
            test_image_path = os.path.join(self.upload_folder, test_image_filename)
            plt.savefig(test_image_path)
            plt.close()

            test_bias_columns[column]["Image"] = test_image_filename
            images.append(test_image_filename)

        final_results = {
            "Train_bias_columns": train_bias_columns,
            "Test_bias_columns": {},
            "Images": images
        }

        final_results = self.describe_data_bias_tests(final_results, test_name='categorical_skewness')

        if isinstance(final_results,dict):
            for _, cols in final_results.items():
                if isinstance(cols,dict):
                    for col, colData in cols.items():
                        if colData["isBiased"]:
                            self.__data_biases["Categorical Skewness"].append(colData["description"][0]) 

        def convert_keys_to_str(obj):
            if isinstance(obj, dict):
                return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_keys_to_str(i) for i in obj]
            else:
                return obj

        return convert_keys_to_str(convert_numpy_types(final_results))

    def calculate_continuous_skewness(self):
        """Calculate continuous skewness for specified columns in training and testing datasets.

        Returns
        -------
        dict
            Results dictionary with skewness values and associated visualizations.

        Raises
        ------
        RuntimeError
            If there is an error during the calculation of continuous skewness.
        """
        continuous_columns = self.__continuous_columns
        if self.train_columns is None:
            raise RuntimeError("No column names provided, unable to evaluate Continuous Skewness.")

        if continuous_columns is None:
            analyze_columns = [col for col in self.__train_df.columns if self.train[col].nunique() > 10]
        else:
            provided_cols = continuous_columns
            if not provided_cols:
                raise RuntimeError("No valid continuous columns provided.")
            for col in provided_cols:
                if col not in self.__train_df.columns or col not in self.__test_df.columns:
                    raise RuntimeError(f"Column '{col}' not found in train/test data.")
                if self.__train_df[col].nunique() <= 10:
                    raise RuntimeError(f"Column '{col}' does not appear to be continuous (≤10 unique values).")
            analyze_columns = provided_cols


        # Check if no columns are available for analysis
        if not analyze_columns:
            return {
                "message": "No continuous columns are available in the dataset for skewness analysis."
            }

        train_skewed_columns = {}
        test_skewed_columns = {}

        for column in analyze_columns:
            train_sk = skew(self.__train_df[column].dropna())
            test_sk = skew(self.__test_df[column].dropna())

            train_img, test_img = None, None
        
            # Generate and store train image
            plt.figure(figsize=(6, 4))
            sns.kdeplot(self.__train_df[column].dropna(), shade=True)
            plt.title(f"Train KDE for {column}")
            plt.xlabel(column)
            plt.ylabel("Density")
            plt.tight_layout()
            train_img = f"{column}_train_kde.png"
            plt.savefig(os.path.join(self.upload_folder, train_img))
            plt.close()

            train_skewed_columns[column] = {
                "skewness": train_sk,
                "image": train_img
            }

            # Generate and store test image
            plt.figure(figsize=(6, 4))
            sns.kdeplot(self.__test_df[column].dropna(), shade=True)
            plt.title(f"Test KDE for {column}")
            plt.xlabel(column)
            plt.ylabel("Density")
            plt.tight_layout()
            test_img = f"{column}_test_kde.png"
            plt.savefig(os.path.join(self.upload_folder, test_img))
            plt.close()

            test_skewed_columns[column] = {
                "skewness": test_sk,
                "image": test_img
            }

        results = {
            "train": train_skewed_columns,
            "test": {}
        }

        results = self.describe_data_bias_tests(results, test_name='continuous_skewness')

        for _, cols in results.items():
            for col, colData in cols.items():
                if colData["isBiased"]:
                    self.__data_biases["Continous Skewness"].append(colData["description"][0])

        def convert_keys_to_str(obj):
            if isinstance(obj, dict):
                return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_keys_to_str(i) for i in obj]
            else:
                return obj
        
        return convert_keys_to_str(results)

    def detect_plot_outliers(self):
        """Detect and visualize outliers for specified columns in training and testing datasets.

        Returns
        -------
        dict
            Results dictionary with outlier detection and visualizations.

        Raises
        ------
        RuntimeError
            If there is an error during outlier detection.
        """
        continuous_columns = self.__continuous_columns
        if self.train_columns is None:
            raise RuntimeError("No column names provided, unable to detect outliers.")

        # Determine columns to analyze
        if continuous_columns is None:
            analyze_columns = [col for col in self.__train_df.columns if self.__train_df[col].nunique() > 20]
        else:
            provided_cols = continuous_columns
            if not provided_cols:
                raise RuntimeError("No valid continuous columns provided.")
            for col in provided_cols:
                if col not in self.__train_df.columns or col not in self.__test_df.columns:
                    raise RuntimeError(f"Column '{col}' not found in train/test data.")
                if self.__train_df[col].nunique() <= 10:
                    raise RuntimeError(f"Column '{col}' not continuous (\u226410 unique values) for outlier detection.")
            analyze_columns = provided_cols

        # Check if there are no continuous columns to analyze
        if not analyze_columns:
            return {"message": "No continuous columns are available in the dataset for outliers detection."}

        def find_outliers(data, column):
            Q1 = data[column].quantile(0.25)
            Q3 = data[column].quantile(0.75)
            IQR = Q3 - Q1
            lower_bound = Q1 - 1.5 * IQR
            upper_bound = Q3 + 1.5 * IQR
            outliers = data[(data[column] < lower_bound) | (data[column] > upper_bound)].index.tolist()
            return outliers

        def plot_outliers(data, column, dataset_name):
            plt.figure(figsize=(6, 4))
            sns.violinplot(x=data[column], color='skyblue')
            plt.title(f'Outliers in {dataset_name} Column: {column}')
            plt.xlabel(column)
            plt.tight_layout()
            img_filename = f"{column}_{dataset_name.lower()}_outliers.png"
            img_path = os.path.join(self.upload_folder, img_filename)
            plt.savefig(img_path)
            plt.close()
            return img_filename

        results = {"train": {}, "test": {}}

        for column in analyze_columns:
            train_outliers = find_outliers(self.__train_df, column)
            if train_outliers:
                train_img = plot_outliers(self.__train_df, column, "Train")
                results["train"][column] = {
                    "outliers": train_outliers,
                    "count": len(train_outliers),
                    "image": train_img
                }


            test_outliers = find_outliers(self.__test_df, column)
            if test_outliers:
                test_img = plot_outliers(self.__test_df, column, "Test")
                results["test"][column] = {
                    "outliers": test_outliers,
                    "count": len(test_outliers), 
                    "image": test_img
                }

        # Add descriptions to results
        if results["train"]:
            self.describe_data_bias_tests(results=results["train"], dataset=self.__train_df, test_name='outliers')
        if results["test"]:
            self.describe_data_bias_tests(results=results["test"], dataset=self.__test_df, test_name='outliers')


        results["test"] = {}

        # Return results or a message if no outliers were found
        if not results["train"] and not results["test"]:
            return {"message": "No outliers detected in the provided continuous columns. No Biasness exists."}
        
        for _, cols in results.items():
            for col, colData in cols.items():
                if colData["isBiased"]:
                    self.__data_biases["Outlier Detection"].append(colData["description"][0])
        
        def convert_keys_to_str(obj):
            if isinstance(obj, dict):
                return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_keys_to_str(i) for i in obj]
            else:
                return obj

        return convert_keys_to_str(results)

    def calculate_disparate_impact(self):
        """Perform disparate impact analysis for specified categorical columns.

        Returns
        -------
        dict
            Results dictionary with Disparate Impact ratios and visualizations.

        Raises
        ------
        RuntimeError
            If there is an error during the disparate impact analysis.
        """

        categorical_columns = self.__categorical_columns
        if self.train_columns is None:
            raise RuntimeError("No column names provided, unable to analyze Disparate Impact.")

        # Assume the 'target_column' is the last column of the dataset
        target_column = self.__train_df.columns[-1]

        # Determine columns to analyze
        if categorical_columns is None:
            default_cols = ["age", "gender", "race"]
            analyze_columns = [c for c in self.train_columns if any(dc in c.lower() for dc in default_cols)]
            if not analyze_columns:
                raise RuntimeError("No default sensitive columns (age, gender, race) found for disparate impact.")
        else:
            provided_cols = categorical_columns
            if not provided_cols:
                raise RuntimeError("No valid categorical columns provided for disparate impact.")
            for col in provided_cols:
                if col not in self.__train_df.columns or col not in self.__test_df.columns:
                    raise RuntimeError(f"Column '{col}' not found in train/test data.")
            analyze_columns = provided_cols

        for col in analyze_columns:
            if self.__train_df[col].nunique() > 10:
                raise RuntimeError(f"Column '{col}' has more than 10 unique values, not suitable for disparate impact analysis.")

        def compute_di_ratios(data, column, target_col):
            """
            Computes DI ratios for each outcome and subgroup.
            For each outcome:
            - Identify the subgroup with the highest proportion for that outcome as the reference group.
            - Calculate the DI for all subgroups relative to that reference group.

            Returns a nested dictionary: { outcome_value: { subgroup: DI_ratio, ... }, ... }
            """
            results = {}
            all_outcomes = data[target_col].unique()
            group_counts = data.groupby([column, target_col]).size().unstack(fill_value=0)

            for outcome in all_outcomes:
                results[outcome] = {}
                if outcome not in group_counts.columns:
                    continue

                subgroup_total_counts = group_counts.sum(axis=1)  
                outcome_proportions = group_counts[outcome] / subgroup_total_counts
                ref_group = outcome_proportions.idxmax()
                ref_prop = outcome_proportions[ref_group]

                for subgroup in group_counts.index:
                    subgroup_count = group_counts.at[subgroup, outcome]
                    subgroup_total = group_counts.loc[subgroup].sum()
                    subgroup_prop = (subgroup_count / subgroup_total) if subgroup_total > 0 else 0

                    if ref_prop > 0:
                        di_ratio = subgroup_prop / ref_prop
                    else:
                        di_ratio = 0.0 

                    results[outcome][subgroup] = di_ratio

            return results

        def plot_di_ratios(ratios_dict, column_name, dataset_name):
            """
            Plots DI ratios for each outcome and subgroup.
            Saves the plot as an image in the 'uploads' directory.
            Returns the image filename.
            """
            plot_data = []
            for outcome_val, subg_dict in ratios_dict.items():
                for subgroup_val, ratio_val in subg_dict.items():
                    plot_data.append({
                        "Subgroup": str(subgroup_val),
                        "Outcome": str(outcome_val),
                        "DI Ratio": ratio_val
                    })

            if not plot_data:
                return None  

            df_plot = pd.DataFrame(plot_data)
            plt.figure(figsize=(8, 5))
            sns.set_style("whitegrid")
            
            ax = sns.barplot(
                data=df_plot, 
                x="Subgroup", 
                y="DI Ratio", 
                hue="Outcome", 
                palette=["#ADD8E6", "#90EE90", "#FFC0CB", "#FFFFE0"]
            )

            ax.axhline(y=0.2, color="red", linestyle="--", label="0.2 threshold")
            ax.set_title(f"Disparate Impact Ratios: {column_name} ({dataset_name})", fontsize=13, pad=12)
            ax.set_xlabel("Subgroups", fontsize=11)
            ax.set_ylabel("DI Ratio", fontsize=11)
            plt.legend()
            plt.tight_layout()

            img_filename = f"{column_name}_{dataset_name.lower()}_disparate_impact.png"
            plt.savefig(os.path.join(self.upload_folder, img_filename))
            plt.close()
            return img_filename

        train_disparate_impact = {}
        test_disparate_impact = {}
        images = []

        # Process each column for both train and test datasets
        for col in analyze_columns:
            train_ratios = compute_di_ratios(self.__train_df, col, target_column)
            train_disparate_impact[col] = train_ratios

            train_img = plot_di_ratios(train_ratios, col, "Train")
            if train_img:
                images.append(train_img)

            test_ratios = compute_di_ratios(self.__test_df, col, target_column)
            test_disparate_impact[col] = test_ratios

            test_img = plot_di_ratios(test_ratios, col, "Test")
            if test_img:
                images.append(test_img)

        results = {
            "train_disparate_impact": train_disparate_impact,
            "test_disparate_impact": {},
            "images": images
        }

        results = self.describe_data_bias_tests(results, test_name='disparate_impact')
        
        for dataset, cols in results.items():
            if dataset in ["train_disparate_impact", "test_disparate_impact"]:  
                for col, colData in cols.items():
                    if colData["isBiased"]:
                        dsc = colData["description"][-1]
                        self.__data_biases["Disparate Impact"].append(dsc[12:])
 

        def convert_keys_to_str(obj):
            if isinstance(obj, dict):
                return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_keys_to_str(i) for i in obj]
            else:
                return obj

        return convert_keys_to_str(results)

    def compare_train_test(self):
        """Compare train and test datasets for categorical bias using Chi-Square tests.

        Returns
        -------
        dict
            Results dictionary containing Chi-Square test results and visualizations.

        Raises
        ------
        RuntimeError
            If there is an error during the comparison of train and test datasets.
        """
        default_sensitive_cols=None
        categorical_columns = self.__categorical_columns
        if self.train_columns is None:
            raise RuntimeError("No column names provided, unable to compare train and test datasets.")

        # Determine columns to analyze
        if categorical_columns is None:
            default_cols = default_sensitive_cols if default_sensitive_cols is not None else ["age", "gender", "race"]
            analyze_columns = [c for c in self.__train_columns if any(dc in c.lower() for dc in default_cols)]
            if not analyze_columns:
                raise RuntimeError("No default sensitive columns (age, gender, race) found for comparison.")
        else:
            provided_cols = categorical_columns
            if not provided_cols:
                raise RuntimeError("No valid categorical columns provided for comparison.")
            for col in provided_cols:
                if col not in self.__train_df.columns or col not in self.__test_df.columns:
                    raise RuntimeError(f"Column '{col}' not found in train/test data.")
            analyze_columns = provided_cols

        # Ensure columns are categorical (≤10 unique subgroups)
        for col in analyze_columns:
            if self.__train_df[col].nunique() > 10:
                raise RuntimeError(
                    f"Column '{col}' has more than 10 unique values, likely continuous. "
                    "Cannot evaluate categorical bias comparison."
                )

        results = {}

        if not os.path.exists(self.upload_folder):
            os.makedirs(self.upload_folder)

        for column in analyze_columns:
            train_counts = self.__train_df[column].value_counts(normalize=True).rename('train')
            test_counts = self.__test_df[column].value_counts(normalize=True).rename('test')
            combined_counts = pd.concat([train_counts, test_counts], axis=1).fillna(0)

            contingency_table = pd.concat([
                self.__train_df[column].value_counts(),
                self.__test_df[column].value_counts()
            ], axis=1).fillna(0)
            contingency_table.columns = ['train', 'test']

            # Perform Chi-Square test
            chi2, p, dof, ex = chi2_contingency(contingency_table)


            combined_counts = combined_counts.reset_index().rename(columns={'index': column})
            melted = combined_counts.melt(id_vars=column, var_name='Dataset', value_name='Proportion')
            sns.set(style="whitegrid")
            plt.figure(figsize=(8, 6))
            sns.barplot(
                data=melted,
                x=column,
                y='Proportion',
                hue='Dataset',
                palette='muted'
            )
            plt.title(f'Proportion Comparison for {column}\nChi-Square p-value: {p:.4f}')
            plt.ylabel('Proportion')
            plt.xlabel(column)
            plt.legend(title='Dataset')

            # Rotate x-axis labels if there are many categories
            if melted[column].nunique() > 5:
                plt.xticks(rotation=45)
            plt.tight_layout()

            graph_filename = f"{column}_chi_square_comparison.png"
            graph_path = os.path.join(self.upload_folder, graph_filename)
            plt.savefig(graph_path)
            plt.close()

            results[column] = {
                "p_value": p,
                "chi2_stat": chi2,
                "degrees_of_freedom": dof,
                "expected_freq": ex.tolist(),  
                "image": graph_filename,
            }
        self.describe_data_bias_tests(results, test_name='compare_train_test')
        for col, colData in results.items():
            if colData["isBiased"]:
                self.__data_biases["Compare Train Test"].append(colData["description"][0])

        def convert_keys_to_str(obj):
            if isinstance(obj, dict):
                return {str(k): convert_keys_to_str(v) for k, v in obj.items()}
            elif isinstance(obj, list):
                return [convert_keys_to_str(i) for i in obj]
            else:
                return obj

        return convert_keys_to_str(results)