"""
EthicalEvaluator Module
=======================

This module provides the `EthicalEvaluator` class, which orchestrates state-based AI workflows 
focused on fairness, ethical compliance, and generating recommendations based on user input 
and evaluations. It integrates with various components such as PDF processing, Large Language Models (LLMs), 
and real-time communication via SocketIO to facilitate interactive evaluations and feedback.

Dependencies:
- os
- json
- requests
- typing (Literal)
- typing_extensions (TypedDict)
- langgraph.graph (StateGraph, START, END)
- langchain_core.messages (AnyMessage)
- base64
- configparser
- concurrent.futures (ThreadPoolExecutor, as_completed)
- langgraph.types (Command)
- IPython.display (Image, display)
- threading
- PDFProcessor
- LLMCaller
"""



import os
import json
import requests
from typing import Literal
from typing_extensions import TypedDict
from langgraph.graph import StateGraph, START, END
from langchain_core.messages import AnyMessage
import base64
import configparser
from concurrent.futures import ThreadPoolExecutor, as_completed
from langgraph.types import Command
from IPython.display import Image, display
import threading
from PDFProcessor import PDFProcessor
from LLMCaller import LLMCaller
import gc

# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')




class AgentState(TypedDict):
    """
    Defines the structure of the agent state for the EthicalEvaluator.

    Attributes
    ----------
    message : AnyMessage
        The current message in the conversation.
    summary : str
        The generated summary from the PDF report.
    goals : list
        List of goals derived from detected biases and ethical guidelines.
    categories : list
        List of goals with assigned weights based on their relevance to ethical guidelines.
    partial_summaries : list
        List of partial summaries generated from each page of the PDF.
    questions : list
        List of questions generated based on the goals.
    answers : list
        List of user answers to the questions.
    curr_ques : int
        Index of the current question being asked.
    followup_ques : list
        List of follow-up questions generated when necessary.
    followup : bool
        Flag indicating whether a follow-up question is needed.
    followup_count : int
        Counter for the number of follow-up questions.
    evaluations : list
        List of evaluations based on user answers.
    compliance_score : list
        List containing the compliance score calculated from evaluations.
    recommendations : list
        List of actionable recommendations generated based on evaluations.
    biases : dict
        Dictionary containing detected biases from various tests.
    """
    message: AnyMessage
    summary: str
    goals: list
    categories: list
    partial_summaries: list
    questions: list
    answers: list
    curr_ques: int
    followup_ques: list
    followup: bool
    followup_count: int
    evaluations: list
    compliance_score: list
    recommendations: list
    biases: dict

class EthicalEvaluator():
    """
    EthicalEvaluator Class
    ======================
    
    Orchestrates state-based AI workflows, handling fairness, ethical compliance, and 
    generating recommendations based on user input and evaluations. It processes PDF reports, 
    interacts with Large Language Models (LLMs) for summarization and analysis, manages 
    real-time communication via SocketIO, and maintains the state of the evaluation process 
    through a state graph.

    Attributes
    ----------
    __llm_caller : LLMCaller
        Instance responsible for interacting with the LLM.
    __pdf_processor : PDFProcessor
        Instance responsible for processing PDF reports.
    socketio : SocketIO, optional
        The SocketIO instance for real-time communication.
    room_id : str, optional
        Unique identifier for the client room in SocketIO communication.
    graph : StateGraph
        The compiled state graph defining the workflow.
    pending_questions : dict
        Tracks pending questions awaiting user responses.
    lock : threading.Lock
        Ensures thread-safe operations for shared resources.
    """
    def __init__(self, socketio_instance=None, room_id=None, report_path=None):
        """
        Initializes the EthicalEvaluator with necessary components and sets up the state graph.

        Parameters
        ----------
        socketio_instance : SocketIO instance, optional
            Instance of SocketIO for real-time communication (default is None).
        room_id : str, optional
            Room identifier for client communication (default is None).
        report_path : str, optional
            Path to the PDF report to be processed (default is None).

        Raises
        ------
        ValueError
            If there is an error initializing the evaluator with the provided files.
        """
        self.__llm_caller = LLMCaller()

        self.__pdf_processor = PDFProcessor(pdf_file_path=report_path)
        self.__pdf_processor.process_pdf()

        self.socketio = socketio_instance  
        self.room_id = room_id  

        rooms = self.socketio.server.manager.rooms

        graph = StateGraph(AgentState)
        
        
        graph.add_node('PDFSummarizerAgent', self.__generate_summary)
        graph.add_node('GoalGeneratorAgent', self.__generate_goals)
        graph.add_node('WeightAssignerAgent', self.__assign_weights)
        graph.add_node('QuestionGeneratorAgent', self.__generate_questions)
        graph.add_node('Interviewer', self.__run_interview)
        graph.add_node('CriticAgent', self.__criticise)
        graph.add_node('FollowupQuestionAgent', self.__generate_followup)
        graph.add_node('ComplianceCalculatorAgent', self.__calculate_compliance_score)
        graph.add_node('RecommendationAgent', self.__generate_recommendations)

        graph.add_edge(START, 'PDFSummarizerAgent')
        graph.add_edge('PDFSummarizerAgent', 'GoalGeneratorAgent')
        graph.add_edge('GoalGeneratorAgent', 'WeightAssignerAgent')
        graph.add_edge('WeightAssignerAgent', 'QuestionGeneratorAgent')
        graph.add_edge('QuestionGeneratorAgent', 'Interviewer')
        graph.add_edge('FollowupQuestionAgent',  'Interviewer')
        graph.add_edge('ComplianceCalculatorAgent', 'RecommendationAgent')
        graph.add_edge('RecommendationAgent', END)
        
        self.graph = graph.compile()

        # Initialize a dictionary to keep track of pending questions per room
        self.pending_questions = {}
        self.lock = threading.Lock()

    def handle_user_answer(self, room_id, answer):
        """
        Processes user answers for pending questions in a thread-safe manner.

        Parameters
        ----------
        room_id : str
            Identifier for the client's room.
        answer : str
            User's response to the pending question.
        """
        with self.lock:
            print(f"[LangGraph.handle_user_answer] Received answer for room_id: {room_id}, answer: {answer}")
            if room_id in self.pending_questions:
                self.pending_questions[room_id]['answer'] = answer
                self.pending_questions[room_id]['event'].set()
                print(f"[LangGraph.handle_user_answer] Answer set and event triggered for room_id: {room_id}")
            else:
                print(f"[LangGraph.handle_user_answer] No pending question for room_id: {room_id}. Ignoring answer.")

    def draw_graph(self):
        """
        Visualizes the state graph using IPython's display functionalities.
        """
        display(Image(self.graph.get_graph().draw_mermaid_png()))
         
    def __encode_image_to_base64(self, image_path):
        """
        Encodes an image file to a base64 string.

        Parameters
        ----------
        image_path : str
            Path to the image file.

        Returns
        -------
        str
            Base64 encoded string of the image.

        Raises
        ------
        FileNotFoundError
            If the image file does not exist.
        """
        with open(image_path, 'rb') as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')
        
    def __get_partial_summary(self, image_base64):
        """
        Generates a partial summary for a given image using the LLM.

        Parameters
        ----------
        image_base64 : str
            Base64 encoded string of the image.

        Returns
        -------
        str
            Partial summary generated by the LLM.

        Raises
        ------
        ValueError
            If the LLM response does not contain the expected content.
        """
        
        partial_summary_system_prompt = config['prompts']['partial_summary_system_prompt']

        payload_messages = [{"role": "system", "content": partial_summary_system_prompt},
                            {"role": "user", "images": [image_base64]}]
        print("Send partial summary request to api.")
        data = self.__llm_caller.get_response(model=config['modelpaths']['llama'], payload_messages=payload_messages)
        print("Get partial summary response from api.")

        if 'message' in data and 'content' in data['message']:
            partial_summaries = data['message']['content']
            return partial_summaries
        else:
            print("No summary content in response JSON.")
            return ""
               
    def __get_final_summary(self, partial_summaries):
        """
        Combines partial summaries into a final comprehensive summary.

        Parameters
        ----------
        partial_summaries : list
            List of partial summaries generated from each page of the PDF.

        Returns
        -------
        str or None
            Final summary combining all partial summaries, or None if generation fails.
        """

        partial_sums = {"\n\n".join(
                                        [f"Page {i+1} Summary:\n{summary}" for i, summary in enumerate(partial_summaries)]
                                    )}
        
        combination_system_prompt = config['prompts']['combination_system_prompt']
        user_message_content = f"""Here are the partial summaries: {partial_sums}. \n Here are the ethical guidelines: {config['guidelines']['ethical_guidelines']}"""

        payload_messages = [
            {"role": "system", "content": combination_system_prompt},
            {"role": "user", "content": user_message_content},
        ]

        data = self.__llm_caller.get_response(model=config['modelpaths']['llama'], payload_messages=payload_messages)

        if 'message' in data and 'content' in data['message']:
            print('Final Summary Obtained')
            
            if self.socketio and self.room_id:
                self.socketio.emit('bot_message', {'message': "Final Summary Obtained"}, room=self.room_id)
                self.socketio.sleep(20)
            return data['message']['content']
        else:
            print("No final summary content in response JSON.")
            return None

    def __generate_summary(self, max_workers=20):
        """
        Processes PDF images in parallel to generate a complete summary.

        Parameters
        ----------
        max_workers : int, optional
            Maximum number of threads for parallel processing (default is 20).

        Returns
        -------
        dict or None
            Dictionary containing the final and partial summaries, or None if processing fails.
        """

        images_base64 = []
        if not os.path.exists('images'):
            if self.socketio and self.room_id:
                self.socketio.emit('bot_message', {'message': "Folder image path does not exist:"}, room=self.room_id)
                self.socketio.sleep(20)
            print(f"Folder path does not exist: {'images'}")
            return None
  
        for filename in sorted(os.listdir('images')):
            if filename.lower().endswith((".png", ".jpg", ".jpeg")):
                full_path = os.path.join('images', filename)
                images_base64.append(self.__encode_image_to_base64(full_path))

        if not images_base64:
            if self.socketio and self.room_id:
                self.socketio.emit('bot_message', {'message': "No images found in the specified folder."}, room=self.room_id)
                self.socketio.sleep(20)
            print("No images found in the specified folder.")
            return None
        

        partial_summaries = []
        with ThreadPoolExecutor() as executor:
            future_to_image = {
                executor.submit(self.__get_partial_summary, img_b64): idx
                for idx, img_b64 in enumerate(images_base64)
            }
            idx = 0
            for future in as_completed(future_to_image):
                try:
                    summary = future.result()
                    message = f"Processed image {idx+1}/{len(images_base64)}: Partial summary obtained."
                    print(message)
                    idx = idx+1
                    if self.socketio and self.room_id:
                        self.socketio.emit('bot_message', {'message': message}, room=self.room_id)
                        self.socketio.sleep(20)

                    partial_summaries.append((idx, summary))
                except Exception as e:
                    print(f"Exception occurred while processing image {idx+1}: {str(e)}")
                    partial_summaries.append((idx, ""))

        partial_summaries_sorted = [summary for idx, summary in sorted(partial_summaries, key=lambda x: x[0])]
        print(partial_summaries_sorted)

        print("Combining partial summaries into a final summary...")

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Combining partial summaries into a final summary"}, room=self.room_id)
            self.socketio.sleep(20)


        final_summary = 'None'
        # final_summary = self.__get_final_summary(                     #Remove this to save time
        #     partial_summaries=partial_summaries_sorted
        # )
        
        return {"summary": final_summary, "partial_summaries": partial_summaries_sorted}
    
    def __generate_goals(self, state: AgentState):
        """
        Generates goals based on detected biases and the project summary.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        dict or None
            Dictionary containing the generated goals, or None if generation fails.
        """
        summary = state['partial_summaries']

        system_prompt = """
You are an expert assistant specializing in fairness and ethical AI analysis. Our system performs fairness tests on datasets to detect biases and assesses compliance with ethical guidelines. Your task is to generate actionable goals for improving fairness and ethical compliance, based on the given context.

Instructions:

Step 1: Generate Goals from Detected Biases
1. Analyze the provided 'biases' dictionary, where:
   - Each key represents a fairness test conducted on the dataset.
   - Each value is a list of conclusions derived from that test.
2. For each key (fairness test), generate one actionable goal by combining all the items in the corresponding list into a single meaningful statement.
3. Each goal should:
   - Clearly describe the bias detected, including affected columns, groups, and numerical differences.
   - Highlight the real-world implications of the bias if left unaddressed.
   - Propose that the dataset or system should be improved to align with real-world equity standards.
4. Append the goals to a dictionary in the following format:
   {
     "Goal1": "...",
     "Goal2": "..."
   }

Step 2: Analyze Ethical Guidelines Compliance
1. Take each ethical guideline and analyze whether the project report summary demonstrates compliance.
2. For each guideline:
   - If compliance is explicitly mentioned, skip the guideline.
   - If compliance is unclear or absent, generate a goal that specifies what is missing in the report and what action is needed to ensure compliance.
3. Repeat this process for all ethical guidelines, which will be provided to you.
4. Append these generated goals to the dictionary created in Step 1.

Final Output:
- Return only the combined goals dictionary, structured as follows:
  {
    "Goal1": "...",
    "Goal2": "...",
    "Goal3": "...",
    ...
  }

Input:
- 'summary': A summary of the project report.
- 'biases': A dictionary containing fairness test results.
- 'guidelines': Ethical Guidelines with categories as mentioned.

Output:
- A dictionary containing all generated goals from Step 1 and Step 2.
"""

        generate_goals_prompt = f"""Here is the summary: {summary}. \n Here is the biases dictionary: {state['biases']}\n Here are the ethical guidelines: {config['guidelines']['ethical_guidelines']} \n\n
        """   
        
        payload_messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": generate_goals_prompt},
        ]

        print('Generating goals')
        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Generating goals"}, room=self.room_id)
            self.socketio.sleep(20)

        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages, format='json')
        if 'message' in data and 'content' in data['message']:
            print(data['message']['content'])
            return {'goals': data['message']['content']}
        else:
            print("No final summary content in response JSON.")
            return None

    def __assign_weights(self, state: AgentState):
        """
        Assigns weights to goals based on their relevance to ethical guidelines.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        None

        Modifies
        --------
        state['categories']
            Appends each goal with its assigned weight.
        """
        print('Assigning weights')

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Assigning weights"}, room=self.room_id)
            self.socketio.sleep(20)

        goals = state['goals']
        
        assign_weights_prompt = """ 
You are a critical analyzer. Given the goals and ethical guidelines by the user, your task is to classify to which guideline does each goal belong to.
Once you have done that, you need to use the following matching rules to assign a weight to each goal:
1. If the goal belongs to 'Ensure Fairness', then assign 0.8 to it.
2. If the goal belongs to 'Promote Transparency', then assign 0.5 to it.
3. If the goal belongs to 'Protect Privacy', then assign 1 to it.
4. If the goal belongs to 'Ensure Accountability', then assign 0.3 to it.
5. If the goal belongs to 'Continuously Monitor AI Systems', then assign 0.2 to it.


For example if the goals contain "Goal1": "The summary does not provide details on the demographic representation in the dataset. For example, it mentions identifying some biases related to overrepresentation of certain patient demographics but lacks specifics such as gender, age, or ethnicity distributions, which are critical for assessing model fairness." then 
Return only the goals and their weightages as a dictionary: 
{
The summary does not provide details on the demographic representation in the dataset. For example, it mentions identifying some biases related to overrepresentation of certain patient demographics but lacks specifics such as gender, age, or ethnicity distributions, which are critical for assessing model fairness.: weight assigned to it,
the next goal: weight assigned to it
...
}
Strictly follow the return format as mentioned above.
"""
        user_prompt = f"""Here are the goals: {goals} \n\n and here are the guidelines:
        {config['guidelines']['ethical_guidelines']}"""
        
        payload_messages = [
            {"role": "system", "content": assign_weights_prompt},
            {"role": "user", "content": user_prompt},
        ]

        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages, temp=0)

        if 'message' in data and 'content' in data['message']:
            categories = data['message']['content']
            categories = json.loads(categories)
            for goal in categories:
                state['categories'].append({'Goal': goal,
                                            'Weight': categories[goal]})
            gc.collect()
            return None
        else:
            print("No categories content in response JSON.")
            return None
        
    def __generate_questions(self, state: AgentState):
        """
        Generates questions based on the provided goals.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        dict or None
            Dictionary containing generated questions, or None if generation fails.
        """
        goals = state['goals']

        question_generator_prompt = f"Here are the Goals: {goals} \n\n "
        
        system_promp = """
You are an expert analyzer and assistant. Our system performs various fairness tests on data to detect biases. We then use the project report provided by them to generate a summary. The summary reflects the results of these analyses and highlights ethical gaps or areas requiring improvement, which are defined as "goals."

### Context Awareness
- Assume the project report and documentation have already been thoroughly analyzed.
- The summary already includes the findings and missing elements, and the goals are derived from those findings.
- Any goal represents an area where the project report either lacks information or fails to meet a requirement.

### Task
For each goal:
1. If the goal is yes/no, ensure the generated question captures the full context by referencing the fairness tests, specific metrics, or gaps identified in the summary. Example:
   - Goal: "In the column 'age_group', subgroup 5 was found to have a high TPR of 0.92 compared to subgroup 0 with a TPR of 0.12, resulting in an equalized odds value of 1.4. This suggests potential bias in treatment outcomes for different age groups."
   - Question: "The equalized odds test shows a high value of 1.4 in the column 'age_group' due to a TPR disparity between subgroup 5 (0.92) and subgroup 0 (0.12). Does this disparity reflect real-world conditions? (Yes/No)"
   
2. For goals requiring elaboration, generate a **detailed question** that explicitly references what was identified or missing in the summary. Example:
   - Goal: "It is unclear if the model was validated using external datasets or if periodic audits were performed to detect emerging biases or performance drifts."
   - Question: "The summary indicates that external validation datasets and periodic audits are not explicitly mentioned. What processes are planned to validate the model and monitor its performance after deployment?"

### Output
Return ONLY the questions in a JSON array of objects with the following structure and no additional text:
- `"goal"`: The complete text of the goal.
- `"question"`: The tailored question, explicitly referencing the findings or missing elements.

### Additional Notes
- For yes/no goals, ensure the question restates the finding or missing element and asks whether it reflects real-world conditions.
- For contextual questions, ensure the question explicitly highlights what has been done or is missing, making it clear how the project can address the identified gaps.
"""

        payload_messages = [
            {"role": "system", "content": system_promp},
            {"role": "user", "content": question_generator_prompt},
        ]
        
        
        print('Generating Survey...')

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Generating Survey"}, room=self.room_id)
            self.socketio.sleep(20)

        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages)

        if 'message' in data and 'content' in data['message']:
            questions = data['message']['content']
            questions = json.loads(questions)
            print(questions)
            return {'questions' : questions}

        else:
            print('Cannot generate questions')
            return None

    def __ask_question(self, question, room_id):
        """
        Emits a question to the client and waits for their response.

        Parameters
        ----------
        question : str
            Question to be asked.
        room_id : str
            Identifier for the client's room.

        Returns
        -------
        str
            User's response to the question.
        """
        ques = f'Question: {question}'
        print(f"[LangGraph.ask_question] Preparing to emit question to room_id: {room_id}")
        print(f"[LangGraph.ask_question] Question: {ques}")

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': ques}, room=room_id)
            print(f"[LangGraph.ask_question] Emitted question to room_id: {room_id}")

            self.socketio.sleep(2)

        # Create an Event and store it
        event = threading.Event()
        with self.lock:
            self.pending_questions[room_id] = {'event': event, 'answer': None}
            print(f"[LangGraph.ask_question] Added pending question for room_id: {room_id}")
            print(f"[LangGraph.ask_question] Current pending_questions: {self.pending_questions}")

         # Wait for the answer
        print(f"[LangGraph.ask_question] Waiting for answer from room_id: {room_id}")
        event.wait()
        print(f"[LangGraph.ask_question] Received event set for room_id: {room_id}")

        # Retrieve the answer
        with self.lock:
            user_answer = self.pending_questions[room_id]['answer']
            del self.pending_questions[room_id]
            print(f"[LangGraph.ask_question] Sending answer forward for room_id: {room_id}: {user_answer}")

        print(f"[LangGraph.ask_question] Return: {room_id}: {user_answer}")
        return user_answer
       
    def __run_interview(self, state: AgentState) -> Command[Literal['CriticAgent', 'ComplianceCalculatorAgent']]:
        """
        Conducts the interview by iterating through questions and handling follow-ups.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        Command
            Next agent to transition to, along with any state updates.
        """
        room_id = self.room_id
        followup_ques = state['followup_ques']
        followup = state['followup']

        if followup and len(followup_ques)>0:
            goal = followup_ques[-1]['goal']
            fquestion = followup_ques[-1]['fquestion']
            user_answer = self.__ask_question(question=fquestion, room_id=room_id)
            print(f"[LangGraph.run_interview] Get in followuppp interview: {room_id}: {user_answer}")

            return Command(goto='CriticAgent', update=state['answers'].append({'goal': str(goal),
                                                                               'question': fquestion,
                                                                               'answer':user_answer}))

        else:
            questions = state['questions']
            curr_ques = state['curr_ques']
            if curr_ques < len(questions): 
            
                question_dict = questions[curr_ques]
                # print(question_dict)
                goal = question_dict['goal']
                question = question_dict['question']

                user_answer = self.__ask_question(question=question, room_id=room_id)
                print(f"[LangGraph.run_interview] Get in interview: {room_id}: {user_answer}")

                return Command(goto='CriticAgent', update=state['answers'].append({'goal': goal,
                                                                                    'question': question,
                                                                                    'answer':user_answer}))

            else:
                return Command(goto='ComplianceCalculatorAgent')
    
    def __criticise(self, state: AgentState) -> Command[Literal['Interviewer', 'FollowupQuestionAgent']]:
        """
        Evaluates user responses to questions and determines if follow-up is needed.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        Command
            Next agent to transition to based on evaluation results.
        """
        answers_list = state['answers']
        curr_ques = state['curr_ques']
        followup_count = state['followup_count']

        goal = answers_list[-1]['goal']
        question = answers_list[-1]['question']
        answer = answers_list[-1]['answer']

        critic_agent_prompt = f"""Here is the goal against which the answer is being checked, the question which was asked to the user and the answer they gave: \n
        Goal: {goal}
        Question: {question}
        Answer: {answer}"""
        

        system_prompt = """
Check if an answer follows the goal based on the question. Decide if it is:
Complied: The answer fully supports the goal.
Partially Complied:The answer supports the goal and has little details.
Violated:The answer goes against the goal or is irrelevant and has aboslutely no details..

Rules:
1. If details are missing but the answer supports the goal, mark it as **Partially Complied**.
   - Set "followup" to 0 because no follow-up is needed for partial compliance.
2. If the answer is against the goal or irrelevant, mark it as **Violated**.
   - Set "followup" to 1 because follow-up is needed to clarify.
3. For yes/no questions:
   - Mark "Complied" or "Violated" based on the answer.
   - Set "followup" to 0 (yes/no questions never require follow-up).

Provided only a JSON response with following structure:
[{
  "extent": "complied / partially complied / violated",
  "details": "Explain why you chose this extent.",
  "followup": 0 or 1 (Set to 1 only if more details are needed and the extent is violated. Always 0 for yes/no questions.)
}]
"""
        payload_messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": critic_agent_prompt},
        ]

        print('Criticising')
        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Criticising"}, room=self.room_id)
            self.socketio.sleep(20)

        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages, temp=0.3, format='json')

        if 'message' in data and 'content' in data['message']:
            evaluation = data['message']['content']
            evaluation = json.loads(evaluation)
            state['evaluations'].append({'followup': evaluation['followup'],
                                                                'extent': evaluation['extent'],
                                                                'details': evaluation['details'],
                                                                'question': question,
                                                                'answer': answer,
                                                                'goal': goal})
            if evaluation['followup'] == 1:
                followup_count = followup_count+1
                if followup_count >=2:
                    print('Max followups reached. Moving on anyway.')
                    curr_ques = curr_ques + 1
                    return Command(goto='Interviewer', update={'curr_ques': curr_ques,
                                                                'followup': False,
                                                                'followup_count':0})       

                else:
    
                    return Command (goto='FollowupQuestionAgent', update={'followup_count': followup_count,
                                                                            'followup': True})
            
            else:
                print('No followup needed. Moving on....')
                curr_ques = curr_ques + 1
                return Command(goto='Interviewer', update={'curr_ques': curr_ques,
                                                            'followup': False,
                                                            'followup_count': 0})                                              
                                                                                                                                                    
        return None

    def __generate_followup(self, state: AgentState):
        """
        Generates follow-up questions based on evaluations that require additional details.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        None
        """
        print('Generating follow-up...')

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Generating follow-up"}, room=self.room_id)
            self.socketio.sleep(20)

        evaluations = state['evaluations']

        last_eval = evaluations[-1] if evaluations else {}

        goal = last_eval['goal']
        question = last_eval['question']
        answer = last_eval['answer']

        followup_agent_prompt = f"""Here is the goal against which the answer is being checked, the question which was asked to the user and the answer they gave:
        Goal: {goal},
        Previous Question: {question}
        User Answer: {answer}"""
        
        system_prompt = """
        You are an expert AI and ML analyzer. The user was asked a question to check if a certain goal is being fulfilled or not but their answer has insufficient information to do so. 
        You need to generate a followup question to clarify or gather missing details.
        Provide the output in the following JSON format:
        {
        'goal': Original goal. Simply copy paste it here.
        'fquestion': The new question generated.
        }
        """

        payload_messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": followup_agent_prompt},
        ]

        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages, format = 'json')

        if 'message' in data and 'content' in data['message']:
            followup_question = data['message']['content']
            followup_question = json.loads(followup_question)
            goal = followup_question['goal']
            fquestion = followup_question['fquestion']
            state['followup_ques'].append({'goal': goal,
                                                    'fquestion': fquestion})
            return None
        return None
    
    def __calculate_compliance_score(self, state:AgentState):
        """
        Calculates the compliance score based on evaluations and goal weights.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        None

        Modifies
        --------
        state['compliance_score']
            Appends the calculated compliance score.
        """
        
        print("Calculating Compliance.")

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Calculating Compliance"}, room=self.room_id)

            self.socketio.sleep(20)


        categories = state['categories']
        evaluations = state['evaluations']

        # Convert categories to a dictionary for easier lookup
        category_weights = {item['Goal']: item['Weight'] for item in categories}

        # Define the hierarchy of extents
        extent_priority = {
            'violated': 1,
            'partially complied': 2,
            'complied': 3
        }

        # Group evaluations by goal and select the one with the highest extent
        grouped_evaluations = {}

        for eval in evaluations:
            goal = eval['goal']
            current_extent = eval['extent']
            current_priority = extent_priority.get(current_extent, 0)
            
            if goal not in grouped_evaluations:
                grouped_evaluations[goal] = eval
            else:
                existing_extent = grouped_evaluations[goal]['extent']
                existing_priority = extent_priority.get(existing_extent, 0)
                if current_priority > existing_priority:
                    grouped_evaluations[goal] = eval

        # Calculate deducted points based on the selected evaluations
        deducted_points = 0

        for goal, eval in grouped_evaluations.items():
            weightage = category_weights.get(goal, 0)
            extent = eval['extent']
            
            if extent == 'violated':
                deducted_points += weightage * 20
            elif extent == 'partially complied':
                deducted_points += weightage * 10

        # Calculate maximum possible deduction
        max_deduction = sum(item['Weight'] * 20 for item in categories)

        # Calculate compliance score
        compliance_score = max(0, (deducted_points / max_deduction) * 100)

        print(f"Deducted Points: {deducted_points}")
        print(f"Compliance Score: {compliance_score:.2f}%")
        state['compliance_score'].append(int(compliance_score))
        return None
    
    def __generate_recommendations(self, state:AgentState):
        """
        Generates actionable recommendations based on evaluations.

        Parameters
        ----------
        state : AgentState
            Current state of the agent.

        Returns
        -------
        None
        """
        print('Generating Recommendations...')

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Generating Recommendations"}, room=self.room_id)
            self.socketio.sleep(20)

        evaluations = state['evaluations']
        recommendations = config['guidelines']['recommendations']

        system_prompt = """ 
You are an expert analyzer. An AI engineer has answered a survey about their healthcare AI project, and their answers have been evaluated. You are given:

1. An evaluations list of dictionaries with the following keys:
   - 'goal': The intended goal of the project.
   - 'question': The question asked.
   - 'details': The evaluation details, including flaws or gaps in the answer.
   - 'answer': The engineer's response.

2. A set of ethical guidelines to ensure the project aligns with ethical standards.
3. Reference recommendations provided for inspiration (but do not copy them directly).

Your Task:
- Focus only on evaluations marked as Partially Complied or Violated.
- Use the goal, details, and ethical guidelines to provide specific, actionable recommendations to address the flaws or gaps.
- Do not provide recommendations for evaluations marked as Complied.

Return only the recommendations along with their impact on patient health in a list format without any additional text:

["recommendation1", "recommendation2", ...]
"""

        user_prompt = f"""
Here are the evaluations: {evaluations} \n\n Here are some generic recommendations: {recommendations} \n\n 
Here are the ethical guidelines: {config['guidelines']['ethical_guidelines']}"""
        
        payload_messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]
        
        data = self.__llm_caller.get_response(model=config['modelpaths']['qwen'], payload_messages=payload_messages, temp=0.5)
  
        if 'message' in data and 'content' in data['message']:
            recoms = data['message']['content']
            recoms = json.loads(recoms)
            print(recoms)
            for item in recoms:
                state['recommendations'].append(item)
            
            if self.socketio and self.room_id:
                self.socketio.emit('bot_message', {'message': "Survey Completed Successfully, Thankyou."}, room=self.room_id)
                self.socketio.sleep(10)

            # Disconnect the server
            print("Disconnecting the server...")
            return None

        if self.socketio and self.room_id:
            self.socketio.emit('bot_message', {'message': "Survey Completed Successfully, Thankyou."}, room=self.room_id)
            self.socketio.sleep(10)

        # Disconnect the server
        print("Disconnecting the server...")

        return None