"""
LLMCaller Module
================

This module provides the `LLMCaller` class, which is responsible for interacting with different 
Large Language Model (LLM) APIs. It facilitates sending messages to the LLMs and retrieving 
their responses. The class is designed to handle multiple models by reading their configurations 
from a `config.ini` file and managing API requests accordingly.

Dependencies:
- requests
- configparser
"""

import requests
import configparser

# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')


class LLMCaller:
    """
    LLMCaller Class
    ================
    
    Handles communication with different LLM APIs based on the model specified. It constructs 
    the appropriate payloads, manages retries for failed requests, and returns the responses 
    from the LLMs.

    Attributes
    ----------
    None

    Methods
    -------
    get_response(model, payload_messages, temp=0.1, format=None)
        Sends a request to the specified LLM API and retrieves the response.
    """
    def __init__(self):
        """
        Initializes the LLMCaller instance.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        pass

    def get_response(self, model, payload_messages, temp = 0.1, format=None):
        """
        Sends a request to the specified LLM API and retrieves the response.

        Parameters
        ----------
        model : str
            The identifier for the LLM model to be used (e.g., 'llama', 'qwen').
        payload_messages : list
            A list of message dictionaries to be sent to the LLM. Each message should follow the 
            structure expected by the LLM API, typically containing 'role' and 'content' keys.
        temp : float, optional
            The temperature parameter for the LLM, controlling the randomness of the output 
            (default is 0.1).
        format : str, optional
            The desired response format. If set to 'json', the payload will include the 
            'format' key. This is specific to certain models like 'qwen' (default is None).

        Returns
        -------
        dict or None
            The JSON response from the LLM if the request is successful; otherwise, None.
        """
        payload = { "model": model, 
                    "messages": payload_messages,
                    "stream": False,
                    "temperature": temp,
                    "options": {
                        "num_ctx": 16*1024}}
        
        if model == config['modelpaths']['llama']:
            for _ in range(3):
                try:
                    response = requests.post(config['modelpaths']['llama_api_url'], json=payload)
                    if response.status_code == 200:
                        data = response.json()
                        return data
                    else:
                        print(f"Error: {response.status_code} - {response.text}")

                except Exception as e:
                    print(f"Exception: {str(e)}")
            return None

        if model == config['modelpaths']['qwen']:
            
            if format == 'json':
                payload = {
                "model": config['modelpaths']['qwen'],
                "messages": payload_messages,
                "format": 'json',
                "stream": False,
                "temperature": 0.1,
                "options": {
                    "num_ctx": 16*1024
                }}

                for _ in range(3):
                    try:
                        response = requests.post(config['modelpaths']['qwen_api_url'], json=payload)
                        if response.status_code == 200:
                            data = response.json()
                            return data
                        else:
                            print(f"Error: {response.status_code} - {response.text}")

                    except Exception as e:
                        print(f"Exception: {str(e)}")
                return None

            else:
                payload = {
                "model": config['modelpaths']['qwen'],
                "messages": payload_messages,
                "stream": False,
                "temperature": 0.1,
                "options": {
                    "num_ctx": 16*1024
                }}

                for _ in range(3):
                    try:
                        response = requests.post(config['modelpaths']['qwen_api_url'], json=payload)
                        if response.status_code == 200:
                            data = response.json()
                            return data
                        else:
                            print(f"Error: {response.status_code} - {response.text}")

                    except Exception as e:
                        print(f"Exception: {str(e)}")
                return None