"""
Orchestrator Module
===================

This module provides the `Orchestrator` class, which facilitates bias detection and evaluation
in machine learning models. It includes functionality for uploading and processing data and model
files, running bias detection algorithms, generating bias reports, serving API endpoints for
bias detection results, and integrating with LangGraph for natural language interactions.
"""

import os
import uuid
import joblib
import shutil
import pickle
import logging
import threading
import configparser
from AlgorithmicBiasAnalyzer import AlgorithmicBiasAnalyzer
from DataBiasAnalyzer import DataBiasAnalyzer
from werkzeug.utils import secure_filename
from EthicalEvaluator import EthicalEvaluator, AgentState
from langchain_core.messages import HumanMessage
from ReportGenerator import ReportGenerator
import json


# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

upload_folder = config['filepaths']['upload_folder']    
os.makedirs(upload_folder, exist_ok=True)

# Global event to signal readiness
report_ready_event = threading.Event()

class Orchestrator:
    """
    Orchestrator Class
    ==================

    Manages the workflow for bias detection and evaluation in machine learning projects.
    Handles file uploads, runs bias detection algorithms, generates reports, and integrates
    with LangGraph for interactive evaluations.
    """
    def __init__(self, socketio):
        """
        Initialize the Orchestrator.

        Parameters
        ----------
        socketio : Flask-SocketIO
            Socket.IO instance for real-time communication.
        """
        self.socketio = socketio
        self.__langgraph_agents = {}  
        self.descriptions_and_results = {
            'EQUALIZED_ODDS_RESULTS': None,
            'PREDICTIVE_PARITY_RESULTS': None,
            'EXPLAINABILITY_BIAS_RESULTS': None,
            'TRAIN_TEST_BIAS_RESULTS': None,
            'CATEGORICAL_SKEWNESS_RESULTS': None,
            'CONTINUOUS_SKEWNESS_RESULTS': None,
            'OUTLIERS_RESULTS': None,
            'DISPARATE_IMPACT_RESULTS': None,
            'COMPARE_TRAIN_TEST': None,
            'PROJECT_REPORT': None,
            'BIASES': None,
            'PROJECT_NAME': None
        }
        self.__report_generator = ReportGenerator()
    
    def __run_algo_analyzer_computations(self, algo_analyzer):
        """
        Execute algorithmic bias detection computations.

        Parameters
        ----------
        algo_analyzer : AlgorithmicBiasAnalyzer
            Instance of AlgorithmicBiasAnalyzer.

        Returns
        -------
        None
        """
 
        try:
            eo_results = algo_analyzer.compute_equalized_odds()
            self.descriptions_and_results['EQUALIZED_ODDS_RESULTS'] = eo_results
        except RuntimeError as re:
            self.descriptions_and_results['EQUALIZED_ODDS_RESULTS'] = {"error": str(re)}

        try:
            pp_results = algo_analyzer.evaluate_predictive_parity()
            self.descriptions_and_results['PREDICTIVE_PARITY_RESULTS'] = {"results": pp_results, "images": pp_results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['PREDICTIVE_PARITY_RESULTS'] = {"error": str(re)}

        try:
            exp_results = algo_analyzer.evaluate_feature_importances()
            self.descriptions_and_results['EXPLAINABILITY_BIAS_RESULTS'] = {"results": exp_results, "images": exp_results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['EXPLAINABILITY_BIAS_RESULTS'] = {"error": str(re)}

        try:
            ttb_results = algo_analyzer.evaluate_confusion_matrix()
            self.descriptions_and_results['TRAIN_TEST_BIAS_RESULTS'] = {"results": ttb_results, "images": ttb_results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['TRAIN_TEST_BIAS_RESULTS'] = {"error": str(re)}
    
    def __run_data_analyzer_computations(self, db_analyzer):
        """
        Execute data bias detection computations.

        Parameters
        ----------
        db_analyzer : DataBiasAnalyzer
            Instance of DataBiasAnalyzer.

        Returns
        -------
        None
        """

        try:
            results = db_analyzer.calculate_categorical_skewness()
            self.descriptions_and_results['CATEGORICAL_SKEWNESS_RESULTS'] = {"results": results, "images": results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['CATEGORICAL_SKEWNESS_RESULTS'] = {"error": str(re)}

        try:
            results = db_analyzer.calculate_continuous_skewness()
            self.descriptions_and_results['CONTINUOUS_SKEWNESS_RESULTS'] = {"results": results, "images": results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['CONTINUOUS_SKEWNESS_RESULTS'] = {"error": str(re)}

        try:
            results = db_analyzer.detect_plot_outliers()
            self.descriptions_and_results['OUTLIERS_RESULTS'] = {"results": results, "images": results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['OUTLIERS_RESULTS'] = {"error": str(re)}

        try:
            results = db_analyzer.calculate_disparate_impact()
            self.descriptions_and_results['DISPARATE_IMPACT_RESULTS'] = {"results": results, "images": results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['DISPARATE_IMPACT_RESULTS'] = {"error": str(re)}

        try:
            results = db_analyzer.compare_train_test()
            self.descriptions_and_results['COMPARE_TRAIN_TEST'] = {"results": results, "images": results.get("images", [])}
        except RuntimeError as re:
            self.descriptions_and_results['COMPARE_TRAIN_TEST'] = {"error": str(re)}

    def process_biases(self, project_name, train_file, test_file, model_file, report_file):
        """
        Process uploaded files and perform bias analysis.

        Parameters
        ----------
        project_name : str
            Name of the project.
        train_file : FileStorage
            Uploaded training data file.
        test_file : FileStorage
            Uploaded testing data file.
        model_file : FileStorage, optional
            Uploaded machine learning model file.
        report_file : FileStorage, optional
            Uploaded report file.

        Returns
        -------
        dict or tuple
            A dictionary containing analysis results or a tuple with an error message and status code.
        """
        train_data = train_file
        test_data = test_file
        upload_folder = config['filepaths']['upload_folder']+f'\{project_name}'
        self.descriptions_and_results["PROJECT_NAME"] = project_name

        os.makedirs(upload_folder, exist_ok=True)
    
        if not train_data or not test_data:
            return "Missing required files. Please upload Train & Test files.", 400

        
        # Save files and load data
        train_filename = secure_filename(train_data.filename)
        unique_train_filename = f"{uuid.uuid4().hex}_{train_filename}"
        train_path = os.path.join(upload_folder, unique_train_filename)
        train_data.save(train_path)

        test_filename = secure_filename(test_data.filename)
        unique_test_filename = f"{uuid.uuid4().hex}_{test_filename}"
        test_path = os.path.join(upload_folder, unique_test_filename)
        test_data.save(test_path)

        loaded_model = None
        if model_file and model_file.filename != '':
            model_filename = secure_filename(model_file.filename)
            unique_model_filename = f"{uuid.uuid4().hex}_{model_filename}"
            model_path = os.path.join(upload_folder, unique_model_filename)
            model_file.save(model_path)
            try:
                if model_filename.endswith('.joblib'):
                    loaded_model = joblib.load(model_path)
                elif model_filename.endswith(('.pkl', '.pickle')):
                    with open(model_path, 'rb') as f:
                        loaded_model = pickle.load(f)
            except Exception as e:
                logger.error(f"Error loading model: {e}")
                os.remove(model_path)
                raise RuntimeError(f"Error loading model: {e}")

        if report_file:
            report_filename = secure_filename(report_file.filename)
            unique_report_filename = f"{uuid.uuid4().hex}_{report_filename}"
            report_path = os.path.join(config['filepaths']['UPLOAD_FOLDER']+'/'+self.descriptions_and_results['PROJECT_NAME'], unique_report_filename)
            report_file.save(report_path)
            self.descriptions_and_results['PROJECT_REPORT'] = report_path
        else:
            self.descriptions_and_results['PROJECT_REPORT'] = None

        #Analyzers Intialized
        db_analyzer = DataBiasAnalyzer(train_file=train_path, test_file=test_path, upload_folder=upload_folder)

        algo_analyzer = AlgorithmicBiasAnalyzer(
            train_file=train_path,
            test_file=test_path,
            model=loaded_model,
            categorical_columns=db_analyzer.compute_categorical_columns(),
            upload_folder=upload_folder
        )


        # Run_algorithmic_analyzer_computations
        t1 = threading.Thread(
            target=self.__run_algo_analyzer_computations,
            args=(algo_analyzer,)
        )
        t1.start()
        t1.join()

        # Run_data_analyzer_computations
        t2 = threading.Thread(
            target=self.__run_data_analyzer_computations,
            args=(db_analyzer,)
        )
        t2.start()
        t2.join()


        algo_ques = algo_analyzer.get_algorithmic_biases()
        bias_ques = db_analyzer.get_data_biases()

        self.descriptions_and_results['BIASES'] = algo_ques | bias_ques

        # Ensure the folder exists
        os.makedirs(upload_folder, exist_ok=True)

        # Path to the JSON file
        file_path = os.path.join(upload_folder, 'descriptions_and_results.json')

        # Save the dictionary as a JSON file
        with open(file_path, 'w') as json_file:
            json.dump(self.descriptions_and_results, json_file, indent=4)

        return self.descriptions_and_results

    def send_user_answer(self, room_id, answer):
        """
        Handle user answers for LangGraph agent.

        Parameters
        ----------
        room_id : str
            Identifier for the communication room.
        answer : str
            User's answer to a question.

        Returns
        -------
        None
        """
        agent = self.__langgraph_agents.get(room_id)
        if not agent:
            self.socketio.emit('state_update', {'message': f'No LangGraph agent found for room_id: {room_id}.'}, room=room_id)
            return
        agent.handle_user_answer(room_id, answer)

    def initiate_langgraph(self, room_id):
        """
        Initialize LangGraph agent for a specific room and generate the project report.

        Parameters
        ----------
        room_id : str
            Identifier for the communication room.

        Returns
        -------
        None
        """
        print(self.descriptions_and_results['PROJECT_REPORT'])
        agent = EthicalEvaluator(socketio_instance=self.socketio, room_id=room_id, report_path=self.descriptions_and_results['PROJECT_REPORT'])
        state = AgentState()

        def limit_list_items(data):
            for key, value in data.items():
                if isinstance(value, list) and len(value) > 2:
                    data[key] = value[:1] 

            if len(data)>5:
                 data=dict(list(data.items())[:2])
            return data
        
        biases = limit_list_items(self.descriptions_and_results['BIASES'])
        print(biases)
        self.__langgraph_agents[room_id] = agent
        
        state['biases'] = biases
        state.setdefault('followup', False)
        state.setdefault('followup_ques', [])
        state.setdefault('answers', [])
        state.setdefault('followup_count', 0)
        state.setdefault('partial_summaries', [])
        state.setdefault('curr_ques', 0)
        state.setdefault('evaluations', [])
        state.setdefault('questions', [])
        state.setdefault('summary', '')
        state.setdefault('categories', [])
        state.setdefault('recommendations', [])
        state.setdefault('compliance_score', [])
        state['message'] = [HumanMessage(content="Use advanced reasoning.")]

        agent.graph.invoke(state, config={'recursion_limit': 200})

        categories = state['categories']
        evaluations = state['evaluations']
        compliance_score = state['compliance_score']
        recommendations = state['recommendations']
    
        self.__report_generator.generate_pdf(categories, evaluations, compliance_score, recommendations,self.descriptions_and_results["PROJECT_NAME"])