"""
PDFProcessor Module
===================
 
This module provides the `PDFProcessor` class, which is responsible for converting PDF files into images. 
It leverages the `pdf2image` library to transform each page of a PDF into a separate image file, facilitating 
further processing such as text extraction or summarization.

Dependencies:
- os
- pathlib (Path)
- pdf2image (convert_from_path)
- configparser
"""

import os
from pathlib import Path
from pdf2image import convert_from_path
import configparser


config = configparser.RawConfigParser()
config.read('config.ini')


class PDFProcessor:
    """
    PDFProcessor Class
    ===================
    
    Handles the conversion of PDF documents into image files. This is particularly useful for scenarios where 
    visual content extraction or image-based analysis is required. The class ensures that the output directory 
    is clean before processing and manages the conversion of each PDF page into a separate image file.

    Attributes
    ----------
    __pdf_file_path : str
        The file path to the PDF document to be processed.
    __output_folder : str
        The directory where the converted images will be saved.

    Methods
    -------
    __clean_output_folder()
        Removes all existing files in the output directory to ensure a clean processing environment.
    process_pdf()
        Converts each page of the PDF into an image and saves them in the output directory.
    """
    def __init__(self, pdf_file_path):
        """
        Initializes the PDFProcessor with the specified PDF file path and sets up the output directory.
        
        Parameters
        ----------
        pdf_file_path : str
            Path to the PDF file to be processed.
        
        Raises
        ------
        FileNotFoundError
            If the specified PDF file does not exist.
        """
        self.__pdf_file_path = pdf_file_path
        self.__output_folder = config['filepaths']['pdf_to_image_output_folder']

        # Ensure the output folder exists
        os.makedirs(self.__output_folder, exist_ok=True)

    def __clean_output_folder(self):
        """
        Cleans the output directory by removing all existing files. This ensures that 
        old images do not interfere with the new processing.

        Raises
        ------
        Exception
            If there is an error deleting any of the files in the output directory.
        """
        for file in Path(self.__output_folder).glob("*"):
            try:
                file.unlink()  # Remove the file
                print(f"Deleted: {file}")
            except Exception as e:
                print(f"Error deleting file {file}: {e}")

    def process_pdf(self):
        """
        Processes the PDF file by converting each of its pages into separate image files.
        
        The method performs the following steps:
        1. Cleans the output directory to remove any existing images.
        2. Converts each page of the PDF into an image using `pdf2image`.
        3. Saves each image in the specified output directory with a standardized naming convention.
        
        Returns
        -------
        list or None
            A list of file paths to the saved images if the processing is successful; 
            otherwise, `None` if an error occurs.
        
        Raises
        ------
        Exception
            If there is an issue during the PDF to image conversion process.
        """
        try:
            # Clean the output folder first
            self.__clean_output_folder()

            # Convert PDF to images
            images = convert_from_path(
                self.__pdf_file_path,
                poppler_path=config['filepaths']['poppler_path']
            )
            saved_images = []

            for i, image in enumerate(images):
                # Define the output path for each image
                output_path = Path(self.__output_folder) / f"page_{i+1}.png"
                image.save(output_path, "PNG")
                saved_images.append(str(output_path))
                print(f"Page {i+1} saved as {output_path}")

            return saved_images
        except Exception as e:
            error_message = str(e)
            if "Unable to get page count" in error_message:
                print(f"Error: Unable to process {self.__pdf_file_path} due to unreadable page count. Skipping this file.")
            else:
                print(f"An unexpected error occurred while processing {self.__pdf_file_path}: {e}")
            return None  

