"""
ReportGenerator Module
======================

This module provides the `ReportGenerator` class, which is responsible for generating a comprehensive PDF report 
based on ethical evaluations conducted on an AI project. It utilizes the `reportlab` library to create well-structured 
and styled PDF documents that include sections such as Goals and Weights, Evaluations, Weighted Violation Index, and Recommendations.

Dependencies:
- os
- reportlab.lib (colors, pagesizes, styles, enums)
- reportlab.platypus (Paragraph, Spacer, Table, TableStyle, Frame, PageTemplate, BaseDocTemplate)
- configparser
"""
import os
from reportlab.lib import colors
from reportlab.lib.pagesizes import LETTER, inch
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import (
    Paragraph,
    Spacer,
    Table,
    TableStyle,
    Frame,
    PageTemplate,
    BaseDocTemplate,
)
from reportlab.lib.enums import TA_CENTER, TA_LEFT
from reportlab.pdfgen import canvas
from reportlab.lib.units import inch
import configparser

# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')

class BorderCanvas(canvas.Canvas):
    """
    Custom Canvas class that adds a lighter border to each page of the PDF.
    
    Inherits from ReportLab's `canvas.Canvas` and overrides the `showPage` and `save` methods 
    to draw a light grey rectangle border on every page.
    """
    def __init__(self, *args, **kwargs):
        super(BorderCanvas, self).__init__(*args, **kwargs)
        self.pagesize = LETTER

    def draw_border(self):
        """
        Draws a light grey border around the page.
        """
        width, height = self.pagesize
        self.setStrokeColor(colors.lightgrey)  
        self.setLineWidth(1)  
        self.rect(0.5 * inch, 0.5 * inch, width - inch, height - inch)

    def showPage(self):
        """
        Overrides the showPage method to include the border before moving to the next page.
        """
        self.draw_border()
        super(BorderCanvas, self).showPage()

    def save(self):
        """
        Overrides the save method to ensure the border is drawn on the last page.
        """
        self.draw_border()
        super(BorderCanvas, self).save()

class ReportGenerator:
    """
    ReportGenerator Class
    =====================
    
    Generates a structured PDF report based on ethical evaluations of an AI project. The report includes 
    sections such as Goals and Weights, Evaluations, Weighted Violation Index, and Recommendations. It leverages 
    the `reportlab` library to create visually appealing and well-organized PDF documents.

    Attributes
    ----------
    __filepath : str
        The base file path where the generated PDFs will be saved, as defined in config.ini.

    Methods
    -------
    add_styles():
        Defines and returns the styles used in the PDF report.
    generate_pdf(categories, evaluations, compliance_score, recommendations, project_name):
        Compiles the provided data into a PDF report and saves it to the specified filepath.
    """
    def __init__(self):
        """
        Initializes the ReportGenerator and sets the output filepath from config.
        """
        self.__filepath = config['filepaths']['upload_folder']
        # Ensure the upload_folder exists
        os.makedirs(self.__filepath, exist_ok=True)

    @staticmethod
    def add_styles():
        """
        Defines and returns a set of paragraph styles for use in the PDF report.
        
        Returns
        -------
        dict
            A dictionary of `ParagraphStyle` objects keyed by style names.
        """
        styles = getSampleStyleSheet()

        # Custom styles
        styles.add(ParagraphStyle(
            name='TitleStyle',
            fontSize=22,
            leading=26,
            alignment=TA_CENTER,
            spaceAfter=20,
            textColor=colors.black,
        ))

        styles.add(ParagraphStyle(
            name='HeadingStyle',
            fontSize=16,
            leading=20,
            spaceBefore=15,
            spaceAfter=10,
            textColor=colors.black,
            fontName='Helvetica-Bold',
        ))

        styles.add(ParagraphStyle(
            name='SubHeadingStyle',
            fontSize=14,
            leading=18,
            spaceBefore=10,
            spaceAfter=5,
            textColor=colors.black,
            fontName='Helvetica-Bold',
        ))

        styles.add(ParagraphStyle(
            name='NormalStyle',
            fontSize=11,
            leading=14,
            spaceAfter=10,
            alignment=TA_LEFT,
        ))

        styles.add(ParagraphStyle(
            name='BoldStyle',
            parent=styles['Normal'],
            fontName='Helvetica-Bold',
        ))

        styles.add(ParagraphStyle(
            name='BoldBlue',
            parent=styles['Normal'],
            fontName='Helvetica-Bold',
            textColor=colors.darkblue,
        ))

        return styles

    def generate_pdf(self, categories, evaluations, compliance_score, recommendations, project_name):
        """
        Generates a PDF report based on the provided data and saves it to the specified filepath.
        
        The report includes:
        - Title: "Ethical Analysis"
        - Goals and Weights: Lists each goal with its assigned weight.
        - Evaluations: Details each evaluation, including the question, answer, extent of compliance, and additional details.
        - Weighted Violation Index: Displays the weighted violation index (WVI) as a percentage.
        - Recommendations: Provides actionable recommendations based on the evaluations.
        
        Parameters
        ----------
        categories : list of dict
            Each dictionary contains 'Goal' and 'Weight' keys representing project goals and their respective weights.
        evaluations : list of dict
            Each dictionary contains 'goal', 'question', 'answer', 'extent', and 'details' keys representing user responses and their evaluations.
        compliance_score : list of int
            A list containing a single integer representing the compliance score percentage.
        recommendations : list of str
            A list of actionable recommendations based on the evaluations.
        project_name : str
            The name of the project for which the report is being generated.
        
        Raises
        ------
        ValueError
            If any of the required parameters are missing or improperly formatted.
        """
        # Validate inputs
        if not all([categories, evaluations, compliance_score, recommendations, project_name]):
            raise ValueError("All parameters must be provided and non-empty.")

        if not isinstance(compliance_score, list) or len(compliance_score) != 1:
            raise ValueError("compliance_score must be a list containing a single integer.")

        # Define the project-specific directory
        project_dir = os.path.join(self.__filepath, project_name)
        os.makedirs(project_dir, exist_ok=True)

        # Define the PDF filename
        pdf_filename = 'ethical_analysis_report.pdf'
        pdf_path = os.path.join(project_dir, pdf_filename)

        # Initialize the PDF document
        doc = BaseDocTemplate(
            pdf_path,
            pagesize=LETTER,
            rightMargin=0.75 * inch,
            leftMargin=0.75 * inch,
            topMargin=0.75 * inch,
            bottomMargin=0.75 * inch,
        )

        # Define frame
        frame = Frame(
            doc.leftMargin,
            doc.bottomMargin,
            doc.width,
            doc.height,
            id='normal'
        )

        # Define page template
        template = PageTemplate(id='test', frames=frame)
        doc.addPageTemplates([template])

        # Get styles
        styles = self.add_styles()

        # Story is a list of flowables
        story = []

        # Title
        story.append(Paragraph("Ethical Analysis", styles['TitleStyle']))
        story.append(Spacer(1, 12))

        # Goals and Weights
        story.append(Paragraph("Goals and Weights", styles['HeadingStyle']))
        for idx, category in enumerate(categories, 1):
            goal_text = f"<b>Goal:</b> {category['Goal']}"
            story.append(Paragraph(goal_text, styles['NormalStyle']))

            weight_text = f"<b>Weight:</b> <font color='darkblue'><b>{category['Weight']}</b></font>"
            story.append(Paragraph(weight_text, styles['NormalStyle']))

            story.append(Spacer(1, 12))

        # Evaluations
        story.append(Paragraph("Evaluations", styles['HeadingStyle']))
        for idx, evaluation in enumerate(evaluations, 1):
            eval_heading = f"Evaluation {idx}"
            story.append(Paragraph(eval_heading, styles['SubHeadingStyle']))

            # Determine text color based on compliance extent
            extent_lower = evaluation['extent'].lower()
            if extent_lower == 'complied':
                extent_color = colors.green
            elif extent_lower in ['partially complied', 'partially violated']:
                extent_color = colors.blue
            else:
                extent_color = colors.red

            data = [
                [Paragraph("<b>Question:</b>", styles['BoldStyle']), Paragraph(evaluation['question'], styles['NormalStyle'])],
                [Paragraph("<b>Answer:</b>", styles['BoldStyle']), Paragraph(evaluation['answer'], styles['NormalStyle'])],
                [Paragraph("<b>Extent of Compliance:</b>", styles['BoldStyle']),
                 Paragraph(
                     evaluation['extent'].capitalize(),
                     ParagraphStyle(
                         name='ComplianceStyle',
                         parent=styles['NormalStyle'],
                         textColor=extent_color
                     )
                 )],
                [Paragraph("<b>Details:</b>", styles['BoldStyle']), Paragraph(evaluation['details'], styles['NormalStyle'])]
            ]

            table = Table(data, colWidths=[2 * inch, 4.5 * inch])
            table.setStyle(TableStyle([
                ('VALIGN', (0, 0), (-1, -1), 'TOP'),
                ('BOTTOMPADDING', (0, 0), (-1, -1), 6),
                ('LEFTPADDING', (0, 0), (-1, -1), 0),
                ('RIGHTPADDING', (0, 0), (-1, -1), 0),
            ]))
            story.append(table)
            story.append(Spacer(1, 12))

        # Weighted Violation Index
        story.append(Paragraph("Weighted Violation Index", styles['HeadingStyle']))

        # Fetch the single element from compliance_score list
        wvi_value = compliance_score[0]

        wvi_paragraph = Table(
            [[Paragraph(f"<b>{wvi_value}%</b>", styles['TitleStyle'])]],
            colWidths=[4 * inch]
        )
        wvi_paragraph.setStyle(TableStyle([
            ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
            ('TEXTCOLOR', (0, 0), (-1, -1), colors.black),
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
            ('TOPPADDING', (0, 0), (-1, -1), 10),
            ('BOTTOMPADDING', (0, 0), (-1, -1), 10),
            ('BOX', (0, 0), (-1, -1), 1, colors.black),  
        ]))

        story.append(wvi_paragraph)
        story.append(Spacer(1, 20))

        # Recommendations
        story.append(Paragraph("Recommendations", styles['HeadingStyle']))
        for rec in recommendations:
            story.append(Paragraph(f"\u2022 {rec}", styles['NormalStyle']))
        story.append(Spacer(1, 12))

        # Build PDF with custom canvas for lighter border
        try:
            doc.build(story, canvasmaker=BorderCanvas)
            print(f"PDF generated successfully at: {pdf_path}")
        except Exception as e:
            print(f"An error occurred while generating the PDF: {e}")