import os
import uuid
import logging
from flask import Flask
from Orchestrator import Orchestrator
from flask import Flask, render_template, request, session, jsonify, send_from_directory
from flask_socketio import SocketIO, emit, join_room, leave_room
import logging
import configparser
from werkzeug.utils import secure_filename

# Configure logging
logging.basicConfig(
    level=logging.DEBUG, 
    format="%(asctime)s - %(levelname)s - %(message)s", 
    datefmt="%Y-%m-%d %H:%M:%S")
# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

socketio_logger = logging.getLogger('socketio')
socketio_logger.setLevel(logging.DEBUG)

engineio_logger = logging.getLogger('engineio')
engineio_logger.setLevel(logging.DEBUG)

# Load configuration from config.ini
config = configparser.RawConfigParser()
config.read('config.ini')

# Initialize Flask and SocketIO
app = Flask(__name__)
app.secret_key = 'your_secret_key_here'
socketio = SocketIO(app, ping_timeout=240, ping_interval=120, async_mode="threading")
app.config['SESSION_COOKIE_SAMESITE'] = 'Lax'

upload_folder = config['filepaths']['upload_folder']
os.makedirs(upload_folder, exist_ok=True)
app.config['UPLOAD_FOLDER'] = upload_folder
print(app.config['UPLOAD_FOLDER'] )

# Initialize Orchestrator
orchestrator = Orchestrator(socketio)


# Routes
@app.route('/')
def home():
    session.clear()
    return render_template('web_home.html')

@app.route('/bias')
def bias():
    return render_template("bias_detect_page.html")

#reroutes to page showing all biases
@app.route('/ethical_eval', methods=['GET', 'POST'])
def ethical_eval():
    if request.method == 'POST':
        try:
            results = orchestrator.process_biases(
                project_name=request.form.get("projectName"),
                train_file=request.files.get('trainData'),
                test_file=request.files.get('testData'),
                model_file=request.files.get('modelFile'),
                report_file=request.files.get('reportFile')
            )
            return render_template("ethical_eval_page.html")
        except Exception as e:
            logger.error(f"Error in ethical_eval: {e}", exc_info=True)
            return "An error occurred while processing the request.", 500
    else:
        return render_template("ethical_eval_page.html")

@app.route('/get_equalized_odds_results', methods=['GET'])
def get_equalized_odds_results():
    results = orchestrator.descriptions_and_results['EQUALIZED_ODDS_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify({"results": results}), 200

@app.route('/get_predictive_parity_results', methods=['GET'])
def get_predictive_parity_results():
    results = orchestrator.descriptions_and_results['PREDICTIVE_PARITY_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_explainability_bias_results', methods=['GET'])
def get_explainability_bias_results():
    results = orchestrator.descriptions_and_results['EXPLAINABILITY_BIAS_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_train_test_bias_results', methods=['GET'])
def get_train_test_bias_results():
    results = orchestrator.descriptions_and_results['TRAIN_TEST_BIAS_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_categorical_skewness_results', methods=['GET'])
def get_categorical_skewness_results():
    results = orchestrator.descriptions_and_results['CATEGORICAL_SKEWNESS_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_continuous_skewness_results', methods=['GET'])
def get_continuous_skewness_results():
    results = orchestrator.descriptions_and_results['CONTINUOUS_SKEWNESS_RESULTS']
    if results is None:
        response = jsonify({"error": "No results available."})
        response.status_code = 400
        return response

    response = jsonify(results)
    response.status_code = 200
    return response

@app.route('/get_outliers_results', methods=['GET'])
def get_outliers_results():
    results = orchestrator.descriptions_and_results['OUTLIERS_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_disparate_impact_results', methods=['GET'])
def get_disparate_impact_results():
    results = orchestrator.descriptions_and_results['DISPARATE_IMPACT_RESULTS']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/get_compare_train_test_results', methods=['GET'])
def get_compare_train_test_results():
    results = orchestrator.descriptions_and_results['COMPARE_TRAIN_TEST']
    if results is None:
        return jsonify({"error": "No results available."}), 400
    return jsonify(results), 200

@app.route('/uploads/<filename>', methods=['GET'])
def uploaded_file(filename):
    print("FILENAME::", filename)
    if filename is None:
        return "Filename not provided", 400
    
    print("Image::", app.config['UPLOAD_FOLDER']+ '/' + orchestrator.descriptions_and_results["PROJECT_NAME"])
    return send_from_directory(app.config['UPLOAD_FOLDER']+'/'+orchestrator.descriptions_and_results["PROJECT_NAME"], filename)

#reroutes to chatbot
@app.route('/upload_report', methods=['POST'])
def upload_report():
    if request.method == 'POST':
        try:
            if 'reportFile' not in request.files:
                return jsonify({"error": "No file part in the request."}), 400

            report_file = request.files.get('reportFile')
            print(report_file)
            save_path = None
 
            if report_file and report_file.filename != '':
                allowed_extensions = {'pdf', 'docx'}
                if not ('.' in report_file.filename and report_file.filename.rsplit('.', 1)[1].lower() in allowed_extensions):
                    return jsonify({"error": "Invalid file type. Only PDF and DOCX are allowed."}), 400
 
                filename = secure_filename(report_file.filename)
                save_path = os.path.join(app.config['UPLOAD_FOLDER']+'/'+orchestrator.descriptions_and_results['PROJECT_NAME'], filename)
                orchestrator.descriptions_and_results['PROJECT_REPORT'] = save_path
                report_file.save(save_path)
            elif orchestrator.descriptions_and_results['PROJECT_REPORT'] and os.path.exists(orchestrator.descriptions_and_results['PROJECT_REPORT']):
                save_path = orchestrator.descriptions_and_results['PROJECT_REPORT']
            else:
                return jsonify({"error": "No file selected for upload."}), 400

            # Generate a unique room ID for the user
            room_id = str(uuid.uuid4())
            session['room_id'] = room_id

            return render_template('chatbot_page.html', room_id=room_id, project_name=orchestrator.descriptions_and_results["PROJECT_NAME"])
        except Exception as e:
            return jsonify({"error": f"An error occurred while uploading the file: {str(e)}"}), 500

    return render_template("chatbot_page.html")

@app.route('/get_report_uploaded', methods=['GET'])
def get_report_uploaded():
    report_path = orchestrator.descriptions_and_results['PROJECT_REPORT']
    if report_path and os.path.exists(report_path):
        filename = os.path.basename(report_path)
        file_url = f"/uploads/{filename}"
        return jsonify({"report_path": file_url})
    else:
        return jsonify({"report_path": None})


# SocketIO Events
@socketio.on('frontend_ready')
def handle_frontend_ready(data):
    room_id = data.get('room_id')
    if room_id:
        socketio.start_background_task(orchestrator.initiate_langgraph, room_id)

@socketio.on('connect')
def handle_connect():
    print('Client connected')

@socketio.on('join')
def on_join(data):
    room_id = data['room_id']
    join_room(room_id)
    print(f'Client joined room: {room_id}')

@socketio.on('disconnect')
def handle_disconnect():
    print('Client disconnected')

@socketio.on('user_answer')
def handle_user_answer_event(data):
    room_id = data.get('room_id')
    answer = data.get('answer')
    if room_id and answer:
        orchestrator.send_user_answer(room_id, answer)
    else:
        emit('state_update', {'message': 'Invalid user answer data.'})


def main():
    """Entry point for running the app."""
    socketio.run(app, host='0.0.0.0', port=5000, debug=True)


if __name__ == '__main__':
    main()